from __future__ import annotations

import logging
import queue
import time
from logging.handlers import QueueHandler
from threading import Event
from typing import Callable, List, Optional, Sequence

from smartpipeline.defaults import CONCURRENCY_WAIT
from smartpipeline.error.exceptions import RetryError
from smartpipeline.error.handling import ErrorManager, RetryManager
from smartpipeline.item import Item, Stop
from smartpipeline.stage import BatchStage, ItemsQueue, Stage, StageTypeVar
from smartpipeline.utils import ConcurrentCounter, ProcessCounter

__author__ = "Giacomo Berardi <giacbrd.com>"


def process(
    stage: Stage,
    item: Item,
    error_manager: ErrorManager,
    retry_manager: RetryManager,
) -> Item:
    """
    Execute the :meth:`.stage.Stage.process` method of a stage for an item
    """
    if error_manager.check_critical_errors(item):
        return item
    time1 = time.time()
    # keeping track of caught exceptions
    caught_retryable_exceptions: List[Exception] = []
    while (
        len(caught_retryable_exceptions) <= retry_manager.max_retries
        or retry_manager.max_retries == 0
    ):
        try:
            stage.logger.debug("%s is processing %s", stage, item)
            processed_item = stage.process(item)
            stage.logger.debug("%s has finished processing %s", stage, processed_item)
            # this can't be in a finally, otherwise it would register the `error_manager.handle` time
            processed_item.set_timing(stage.name, time.time() - time1)
            return processed_item
        except retry_manager.retryable_errors as retryable_exc:
            caught_retryable_exceptions.append(retryable_exc)
            if (
                retry_manager.max_retries == 0
            ):  # we have already done an attempt, no more retries
                break  # breaking the loop and attach to the item a RetryError
            time.sleep(
                pow(2, len(caught_retryable_exceptions) - 1) * retry_manager.backoff
            )
        except Exception as e:
            stage.logger.debug("%s has failed processing %s", stage, item)
            item.set_timing(stage.name, time.time() - time1)
            error_manager.handle(e, stage, item)
            return item
    stage.logger.debug(
        "%s has failed processing the %s many times with %s errors",
        stage,
        item,
        len(caught_retryable_exceptions),
    )
    item.set_timing(stage.name, time.time() - time1)
    for rexc in caught_retryable_exceptions:
        error_manager.handle(RetryError().with_exception(rexc), stage, item)
    return item


def process_batch(
    stage: BatchStage,
    items: Sequence[Item],
    error_manager: ErrorManager,
    retry_manager: RetryManager,
) -> List[Optional[Item]]:
    """
    Execute the :meth:`.stage.BatchStage.process_batch` method of a batch stage for a batch of items
    """
    ret: List[Optional[Item]] = [None] * len(items)
    to_process = {}
    for i, item in enumerate(items):
        if error_manager.check_critical_errors(item):
            ret[i] = item
        else:
            stage.logger.debug("%s is going to process %s", stage, item)
            to_process[i] = item
    time1 = time.time()
    # keeping track of caught exceptions
    caught_retryable_exceptions: List[Exception] = []
    while (
        len(caught_retryable_exceptions) <= retry_manager.max_retries
        or retry_manager.max_retries == 0
    ):
        try:
            stage.logger.debug("%s is processing %s items", stage, len(to_process))
            processed = stage.process_batch(list(to_process.values()))
            stage.logger.debug(
                "%s has finished processing %s items", stage, len(to_process)
            )
            spent = (time.time() - time1) / (len(to_process) or 1.0)
            for n, i in enumerate(to_process.keys()):
                item = processed[n]
                item.set_timing(stage.name, spent)
                ret[i] = item
            return ret
        except retry_manager.retryable_errors as retryable_exc:
            caught_retryable_exceptions.append(retryable_exc)
            if (
                retry_manager.max_retries == 0
            ):  # we have already done an attempt, no more retries
                break  # breaking the loop and attach to the item a RetryError
            time.sleep(
                pow(2, len(caught_retryable_exceptions) - 1) * retry_manager.backoff
            )
        except Exception as e:
            stage.logger.debug(
                "%s had failures in processing %s items", stage, len(to_process)
            )
            spent = (time.time() - time1) / (len(to_process) or 1.0)
            for i, item in to_process.items():
                item.set_timing(stage.name, spent)
                error_manager.handle(e, stage, item)
                ret[i] = item
            return ret
    stage.logger.debug(
        "%s has failed in processing %s items many times with %s errors",
        stage,
        len(to_process),
        len(caught_retryable_exceptions),
    )
    spent = (time.time() - time1) / (len(to_process) or 1.0)
    for i, item in to_process.items():
        item.set_timing(stage.name, spent)
        for rexc in caught_retryable_exceptions:
            error_manager.handle(RetryError().with_exception(rexc), stage, item)
        ret[i] = item
    return ret


def stage_runner(
    stage: Stage,
    in_queue: ItemsQueue,
    out_queue: ItemsQueue,
    error_manager: ErrorManager,
    retry_manager: RetryManager,
    terminated: Event,
    has_started_counter: ConcurrentCounter,
    counter: ConcurrentCounter,
    logs_queue: Optional[queue.Queue[logging.LogRecord]],
) -> None:
    """
    Consume items from an input queue, process and put them in an output queue, indefinitely,
    until a termination event is set
    """
    on_multiprocess = isinstance(counter, ProcessCounter)
    if on_multiprocess:
        if logs_queue is not None:
            root_logger = logging.getLogger()
            # only by comparing string of queues we obtain their "original" address
            if not any(
                isinstance(handler, QueueHandler)
                and str(handler.queue) == str(logs_queue)
                for handler in root_logger.handlers
            ):
                handler = QueueHandler(logs_queue)
                root_logger.addHandler(handler)
        # call these only if the stage and the error manager are copies of the original,
        # ergo this executor is running in a child process
        error_manager.on_start()
        stage.on_start()
    has_started_counter += 1
    while True:
        if terminated.is_set() and in_queue.empty():
            if on_multiprocess:
                error_manager.on_end()
                stage.on_end()
            return
        try:
            item = in_queue.get(block=True, timeout=CONCURRENCY_WAIT)
        except queue.Empty:
            continue
        if isinstance(item, Stop):
            out_queue.put(item, block=True)
            in_queue.task_done()
        elif item is not None:
            try:
                item = process(stage, item, error_manager, retry_manager)
            except Exception as e:
                if on_multiprocess:
                    error_manager.on_end()
                    stage.on_end()
                raise e
            else:
                if item is not None:
                    out_queue.put(item, block=True)
                    if not isinstance(item, Stop):
                        counter += 1
            finally:
                in_queue.task_done()


def batch_stage_runner(
    stage: BatchStage,
    in_queue: ItemsQueue,
    out_queue: ItemsQueue,
    error_manager: ErrorManager,
    retry_manager: RetryManager,
    terminated: Event,
    has_started_counter: ConcurrentCounter,
    counter: ConcurrentCounter,
    logs_queue: Optional[queue.Queue[logging.LogRecord]],
) -> None:
    """
    Consume items in batches from an input queue, process and put them in an output queue, indefinitely,
    until a termination event is set
    """
    on_multiprocess = isinstance(counter, ProcessCounter)
    if on_multiprocess:
        if logs_queue is not None:
            root_logger = logging.getLogger()
            # only by comparing string of queues we obtain their "original" address
            if not any(
                isinstance(handler, QueueHandler)
                and str(handler.queue) == str(logs_queue)
                for handler in root_logger.handlers
            ):
                handler = QueueHandler(logs_queue)
                root_logger.addHandler(handler)
        # call these only if the stage and the error manager are copies of the original,
        # ergo this executor is running in a child process
        error_manager.on_start()
        stage.on_start()
    has_started_counter += 1
    while True:
        if terminated.is_set() and in_queue.empty():
            if on_multiprocess:
                error_manager.on_end()
                stage.on_end()
            return
        items: List[Item] = []
        try:
            for _ in range(stage.size):
                item = in_queue.get(block=True, timeout=stage.timeout)
                # give priority to the Stop event item
                if isinstance(item, Stop):
                    out_queue.put(item, block=True)
                elif item is not None:
                    items.append(item)
                in_queue.task_done()
        except queue.Empty:
            if not any(items):
                continue
        if any(items):
            try:
                processed_items = process_batch(
                    stage, items, error_manager, retry_manager
                )
            except Exception as e:
                if on_multiprocess:
                    error_manager.on_end()
                    stage.on_end()
                raise e
            else:
                for final_item in processed_items:
                    if final_item is not None:
                        out_queue.put(final_item, block=True)
                        if not isinstance(final_item, Stop):
                            counter += 1


StageRunner = Callable[
    [
        StageTypeVar,
        ItemsQueue,
        ItemsQueue,
        ErrorManager,
        RetryManager,
        Event,
        ConcurrentCounter,
        ConcurrentCounter,
        Optional[queue.Queue[logging.LogRecord]],
    ],
    None,
]



from __future__ import annotations

from typing import Optional

__author__ = "Giacomo Berardi <giacbrd.com>"


class Error(Exception):
    """
    Base exception type for stages
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._pipeline_stage = None
        self._exception_cause = None

    def set_stage(self, stage: str):
        self._pipeline_stage = stage

    def get_stage(self) -> Optional[str]:
        return getattr(self, "_pipeline_stage", None)

    def with_exception(self, exception: Exception) -> Error:
        """
        Set the original exception (if any) that has generated this error,
        equivalent to `explicit exception chaining <https://www.python.org/dev/peps/pep-3134/#explicit-exception-chaining>`_
        """
        # they are both necessary, in particular the second one is to preserve the cause from child thread exception
        self.__cause__ = exception
        self._exception_cause = exception
        return self

    def get_exception(self) -> Optional[BaseException]:
        """
        Get the original exception (if any) that has generated this error,
        equivalent to the `__cause__ <https://www.python.org/dev/peps/pep-3134/#explicit-exception-chaining>`_ attribute
        """
        return self.__cause__ or self._exception_cause


class SoftError(Error):
    """
    A type of exception which usually only provokes skipping a stage for an item
    """

    pass


class CriticalError(Error):
    """
    A type of exception which usually provokes skipping the whole pipeline for an item
    """

    pass


class RetryError(SoftError):
    """
    A stage processing an input item may raise some Exception for which the stage must try to reprocess the item using
    an exponential backoff retry strategy. After several attempts a RetryError is raised and
    that stage is skipped for that item
    """

    pass



from __future__ import annotations

import logging
import threading
from abc import ABC, abstractmethod
from multiprocessing.managers import SyncManager
from queue import Queue

__author__ = "Giacomo Berardi <giacbrd.com>"

from typing import Any, OrderedDict


def last_key(ordered_dict: OrderedDict[str, Any]) -> str:
    return next(reversed(ordered_dict.keys()))


class ConcurrentCounter(ABC):
    """
    Interface for a counter that is safe for concurrent access
    """

    @abstractmethod
    def __iadd__(self, incr: int) -> ConcurrentCounter:
        return self

    @property
    @abstractmethod
    def value(self) -> int:
        return 0


class ThreadCounter(ConcurrentCounter):
    """
    Thread safe counter
    """

    def __init__(self):
        self._value = 0
        self._lock = threading.Lock()

    def __iadd__(self, incr: int) -> ThreadCounter:
        with self._lock:
            self._value += incr
        return self

    @property
    def value(self) -> int:
        with self._lock:
            return self._value


class ProcessCounter(ConcurrentCounter):
    """
    Process safe counter
    """

    def __init__(self, manager: SyncManager):
        self._value = manager.Value("i", 0)
        self._lock = manager.Lock()

    def __iadd__(self, incr: int) -> ProcessCounter:
        with self._lock:
            self._value.value += incr
        return self

    @property
    def value(self) -> int:
        with self._lock:
            return self._value.value


class LogsReceiver:
    """Read from the queue where sub-processes send their log records and logs them in the main process"""

    def __init__(self, logs_queue: Queue):
        self._logs_queue = logs_queue
        self._thread = None

    def start(self):
        if self._thread is None:

            def _receiver(queue):
                while True:
                    record = queue.get()
                    if record is None:
                        queue.task_done()
                        break
                    logger = logging.getLogger(record.name)
                    logger.handle(record)
                    queue.task_done()

            self._thread = threading.Thread(target=_receiver, args=(self._logs_queue,))
            self._thread.start()

    def stop(self):
        if self._thread is not None:
            self._logs_queue.put(None)
            self._logs_queue.join()
            self._thread.join()
        self._thread = None

    @property
    def queue(self) -> Queue:
        return self._logs_queue



from __future__ import annotations

import logging
import uuid
from typing import Any, Optional, Tuple, Type, Union

from smartpipeline.error.exceptions import CriticalError, SoftError
from smartpipeline.item import Item
from smartpipeline.stage import AliveMixin

__author__ = "Giacomo Berardi <giacbrd.com>"


class ErrorManager:
    """
    Basic error handling of a pipeline, principally manages :class:`.exceptions.SoftError` and :class:`.exceptions.CriticalError` types
    """

    _raise_on_critical = False
    _skip_on_critical = True

    def raise_on_critical_error(self) -> ErrorManager:
        """
        Set the error manager so that if a :class:`.exceptions.CriticalError` (or any un-managed exception)
        is encountered, it raises it "externally", at the pipeline run level, killing the pipeline
        """
        self._raise_on_critical = True
        return self

    def no_skip_on_critical_error(self) -> ErrorManager:
        """
        Change default behaviour of the error manager on :class:`.exceptions.CriticalError`: only skip the stage which raises it,
        like the :class:`.exceptions.SoftError` .
        Valid only if :meth:`.ErrorManager.raise_on_critical_error` is not set
        """
        self._skip_on_critical = False
        return self

    def on_start(self) -> Any:
        """
        Called for a concurrent stage executor in a process (only when multiprocessing concurrency)
        or simply after construction, by the pipeline.
        The error manager in the executor is a copy of the original,
        by overriding this method one can initialize variables specifically for the copies, that is mandatory
        when they are not serializable.
        """
        pass

    def on_end(self) -> Any:
        """
        Called when the pipeline terminates, useful for executing closing operations (e.g. on files)
        """
        pass

    def handle(
        self, error: Exception, stage: AliveMixin, item: Item
    ) -> Optional[CriticalError]:
        """
        Manage an error produced by a stage

        :param error: It can be a generic exception or an error from :mod:`.exceptions` explicitly raised by a stage
        :param stage: Stage which raised the exception during processing
        :param item: Item which raised the exception when processed
        :return: If the handled error results to be critical return the generated :class:`.exceptions.CriticalError`
        :raises Exception: When a :meth:`.ErrorManager.raise_on_critical_error` has been set and the error is critical
        """
        item_error: Union[SoftError, CriticalError]
        if isinstance(error, SoftError):
            item_error = item.add_soft_error(stage.name, error)
        else:
            # any un-managed exception is a potential critical error
            item_error = item.add_critical_error(stage.name, error)
        exc_info = (type(item_error), item_error, item_error.__traceback__)
        self.logger.exception(
            "%s has generated an error on item %s", stage, item, exc_info=exc_info
        )
        if isinstance(item_error, CriticalError):
            exception = self._check_critical(item_error)
            if exception:
                return item_error
        return None

    @property
    def logger(self) -> logging.Logger:
        if getattr(self, "_logger", None) is None:
            self._logger = logging.getLogger(
                f"{self.__class__.__name__}-{str(uuid.uuid4())[:8]}"
            )
        return self._logger

    def _check_critical(
        self, error: CriticalError
    ) -> Optional[Union[BaseException, CriticalError]]:
        """
        Manage a critical error, usually after an item has been processed by a stage

        :return: The exception which caused a critical error if any, otherwise the :class:`.exceptions.CriticalError` itself
        :raises Exception: When a :meth:`.ErrorManager.raise_on_critical_error` has been set
        """
        ex = error.get_exception()
        if self._raise_on_critical:
            raise ex or error
        elif self._skip_on_critical:
            return ex or error
        return None

    def check_critical_errors(self, item: Item) -> Optional[BaseException]:
        """
        Check the critical errors registered for an item
        """
        if item.has_critical_errors():
            for er in item.critical_errors():
                ex = self._check_critical(er)
                if ex:
                    return ex
        return None


class RetryManager:
    """
    This class encapsulate the parameters used to handle the retry strategy in case some kind of error are raise by the stage
    """

    def __init__(
        self,
        retryable_errors: Tuple[Type[Exception], ...] = tuple(),
        max_retries: int = 0,
        backoff: Union[float, int] = 0,
    ):
        """
        :param retryable_errors: tuple of errors types which the retry strategy is applied for
        :param max_retries: maximum number of attempts for a which a stage is run in case of one of the `retryable_errors` is raised during its execution
        :param backoff: weight for the exponential back-off strategy
        """
        self._backoff = backoff
        self._max_retries = max_retries
        self._retryable_errors = retryable_errors

    @property
    def backoff(self) -> float:
        return self._backoff

    @property
    def max_retries(self) -> int:
        return self._max_retries

    @property
    def retryable_errors(self) -> Tuple[Type[Exception], ...]:
        return self._retryable_errors



from __future__ import annotations

import uuid
from typing import Any, Callable, Dict, Generator, KeysView, Union

from smartpipeline.defaults import DATA_SNIPPET_SIZE
from smartpipeline.error.exceptions import CriticalError, SoftError

__author__ = "Giacomo Berardi <giacbrd.com>"


class Item:
    """
    An item containing a unit of data to process through a pipeline.
    Is ia usually generated by the :class:`.stage.Source` subclass and retrieved at the end of a :class:`.pipeline.Pipeline`.
    """

    def __init__(self):
        self._soft_errors = []
        self._critical_errors = []
        self._meta = {}
        self._payload = {}
        self._timings = {}
        self._callback_fun = None
        self._id = None

    def __str__(self) -> str:
        return f"Data item {self.id} with payload {self.data_snippet()}..."

    @property
    def data(self) -> Dict[str, Any]:
        """
        Access to the actual data contained in the item

        :return: A dictionary in which organizing data by fields (recommended to be JSON serializable)
        """
        return self._payload

    # deprecated
    payload = data  # pragma: no cover

    def data_snippet(self, max_size: int = DATA_SNIPPET_SIZE):
        """
        A short string representation of the :attr:`.Item.data` (recommended to override this method)

        :param max_size: Maximum size of the string representation
        """
        return str(self.data)[:max_size]

    def set_metadata(self, field: str, value: Any) -> Item:  # pragma: no cover
        """
        DEPRECATED: Add a metadata, something we want to remember but keep outside the actual data in :attr:`.Item.data`

        :param field: Name of the metadata variable
        """
        self._meta[field] = value
        return self

    def get_metadata(self, field: str) -> Any:  # pragma: no cover
        """
        DEPRECATED: Get a metadata value by its name

        :return: A value or None if the metadata does not exist in the item
        """
        return self._meta.get(field)

    @property
    def metadata(self) -> Dict[str, Any]:
        """
        Get the actual metadata dictionary
        """
        return self._meta

    def set_timing(self, stage_name: str, seconds: float) -> Item:
        """
        Set the time spent by a stage (referenced by its name) for processing the item
        """
        self._timings[stage_name] = seconds
        return self

    def get_timing(self, stage_name: str) -> float:
        """
        Get the time spent by a stage (referenced by its name) for processing the item

        :return: The time in seconds or None if the item has not been processed by the stage
        """
        return self._timings.get(stage_name)

    @property
    def timed_stages(self) -> KeysView[str]:
        """
        Get the stage names for which the item has a process time
        """
        return self._timings.keys()

    @property
    def id(self) -> Any:
        """
        Get the unique identifier of the item.
        It is recommended to override this in order to properly compute it from the :attr:`.Item.data`
        """

        ret = self._payload.get("id")
        if ret is None:
            ret = self._meta.get("id")
            if ret is None:
                if self._id is None:
                    self._id = str(uuid.uuid4())
                return self._id
        return ret

    def set_callback(self, fun: Callable[[Item], Any]):
        """
        Set a function to call after a successful asynchronous execution of a pipeline on the item (through :meth:`.pipeline.Pipeline.process_async`)
        """
        self._callback_fun = fun

    def callback(self) -> Any:
        """
        Call the function set with :meth:`.Item.set_callback`
        """
        if self._callback_fun is not None:
            return self._callback_fun(self)

    def has_soft_errors(self) -> bool:
        """
        True if the item has raised an :class:`.error.exceptions.SoftError` in some stage processing
        """
        return any(self._soft_errors)

    def has_critical_errors(self) -> bool:
        """
        True if the item has raised a :class:`.error.exceptions.CriticalError` or any un-managed exception in some stage processing
        """
        return any(self._critical_errors)

    def soft_errors(self) -> Generator[SoftError, None, None]:
        """
        Iter over :class:`.error.exceptions.SoftError` instances eventually generated in some stage processing
        """
        for e in self._soft_errors:
            yield e

    def critical_errors(self) -> Generator[CriticalError, None, None]:
        """
        Iter over :class:`.error.exceptions.CriticalError` instances or any un-managed exception eventually generated in some stage processing
        """
        for e in self._critical_errors:
            yield e

    def add_soft_error(
        self, stage: str, exception: Union[SoftError, Exception]
    ) -> SoftError:
        """
        *This is called internally by the* :class:`.error.handling.ErrorManager` *when the exception is handled*.
        Add an :class:`.error.exceptions.SoftError` generated in a stage (referenced by its name) for the item.

        :param exception: It can be an :class:`.error.exceptions.SoftError` instance or any exception, which will be encapsulated in an :class:`.error.exceptions.SoftError`
        """
        if type(exception) is not CriticalError:
            if isinstance(exception, SoftError):
                exception.set_stage(stage)
                self._soft_errors.append(exception)
                return exception
            elif isinstance(exception, Exception):
                error = SoftError(str(exception))
                error.with_exception(exception)
                error.set_stage(stage)
                self._soft_errors.append(error)
                return error
        raise ValueError("Add a pipeline SoftError or a generic exception")

    def add_critical_error(
        self, stage: str, exception: Union[CriticalError, Exception]
    ) -> CriticalError:
        """
        *This is called internally by the* :class:`.error.handling.ErrorManager` *when the exception is handled*.
        Add a :class:`.error.exceptions.CriticalError` generated in a stage (referenced by its name) for the item

        :param exception: It can be a :class:`.error.exceptions.CriticalError` instance or any exception, which will be encapsulated in a :class:`.error.exceptions.CriticalError`
        """
        if type(exception) is not SoftError:
            if isinstance(exception, CriticalError):
                exception.set_stage(stage)
                self._critical_errors.append(exception)
                return exception
            elif isinstance(exception, Exception):
                error = CriticalError(str(exception))
                error.with_exception(exception)
                error.set_stage(stage)
                self._critical_errors.append(error)
                return error
        raise ValueError("Add a pipeline CriticalError or a generic exception")


# deprecated
DataItem = Item  # pragma: no cover


class Stop(Item):
    """
    Core "signal" item (*do not use this explicitly*) used in pipelines for determining the end of a flow of items,
    like an event passed through all the stages
    """

    def __str__(self) -> str:
        return f"Stop signal {self.id}"



import copy
import logging
import time
import traceback

import pytest

from tests.utils import (
    BatchErrorStage,
    BatchExceptionStage,
    BatchTextDuplicator,
    BatchTextReverser,
    ErrorStage,
    ExceptionStage,
    RandomTextSource,
    TextDuplicator,
    TextReverser,
    TimeWaster,
    get_pipeline,
)

__author__ = "Giacomo Berardi <giacbrd.com>"

logger = logging.getLogger(__name__)


def _check(items, num, pipeline=None):
    diff = frozenset(range(1, num + 1)).difference(item.data["count"] for item in items)
    assert not diff, "Not found items: {}".format(", ".join(str(x) for x in diff))
    if pipeline:
        assert len(items) == num == pipeline.count


def test_concurrent_run():
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(96))
        .append("reverser0", BatchTextReverser(), concurrency=2)
        .append("reverser1", BatchTextReverser(), concurrency=0)
        .append("reverser2", BatchTextReverser(), concurrency=1)
        .append("duplicator", BatchTextDuplicator(), concurrency=2)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 96, pipeline)
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(81))
        .append("reverser0", BatchTextReverser(), concurrency=2, parallel=True)
        .append("reverser1", BatchTextReverser(), concurrency=1, parallel=True)
        .append("reverser2", BatchTextReverser(), concurrency=0)
        .append("duplicator", BatchTextDuplicator(), concurrency=2, parallel=True)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 81, pipeline)
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(92))
        .append("reverser0", BatchTextReverser(), concurrency=0)
        .append("reverser1", BatchTextReverser(), concurrency=1)
        .append("duplicator", BatchTextDuplicator(), concurrency=0)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 92, pipeline)
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(83))
        .append("reverser0", BatchTextReverser(), concurrency=1, parallel=True)
        .append("reverser1", BatchTextReverser(), concurrency=1, parallel=False)
        .append("duplicator", BatchTextDuplicator(), concurrency=1, parallel=True)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 83, pipeline)
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(101))
        .append("duplicator0", BatchTextDuplicator(), concurrency=0)
        .append("reverser", BatchTextReverser(), concurrency=0)
        .append("duplicator1", BatchTextDuplicator(), concurrency=0)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 101, pipeline)
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(100))
        .append("reverser", BatchTextReverser(), concurrency=0)
        .append("duplicator0", BatchTextDuplicator(), concurrency=0)
        .append("duplicator1", BatchTextDuplicator(check_batch=True), concurrency=1)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 100, pipeline)


def test_queue_sizes():
    pipeline = (
        get_pipeline(max_queues_size=1)
        .set_source(RandomTextSource(93))
        .append("reverser0", BatchTextReverser(), concurrency=2)
        .append("reverser1", BatchTextReverser(), concurrency=0)
        .append("reverser2", BatchTextReverser(), concurrency=1)
        .append("duplicator", BatchTextDuplicator(), concurrency=2)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 93, pipeline)
    pipeline = (
        get_pipeline(max_queues_size=1)
        .set_source(RandomTextSource(79))
        .append("reverser0", BatchTextReverser(), concurrency=2, parallel=True)
        .append("reverser1", BatchTextReverser(), concurrency=1, parallel=True)
        .append("reverser2", BatchTextReverser(), concurrency=0)
        .append("duplicator", BatchTextDuplicator(), concurrency=2, parallel=True)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 79, pipeline)
    pipeline = (
        get_pipeline(max_queues_size=2)
        .set_source(RandomTextSource(131))
        .append("reverser0", BatchTextReverser(), concurrency=2, parallel=True)
        .append("waster1", TimeWaster(0.02), concurrency=2, parallel=True)
        .append("reverser1", BatchTextReverser(), concurrency=0)
        .append("waster2", TimeWaster(0.03), concurrency=2)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 131, pipeline)
    pipeline = (
        get_pipeline(max_queues_size=0)
        .set_source(RandomTextSource(4))
        .append("reverser0", BatchTextReverser(), concurrency=2)
        .append("reverser1", BatchTextReverser(), concurrency=0)
        .append("reverser2", BatchTextReverser(), concurrency=1)
        .append("duplicator", BatchTextDuplicator(), concurrency=2)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 4, pipeline)
    pipeline = (
        get_pipeline(max_queues_size=0)
        .set_source(RandomTextSource(11))
        .append("reverser0", BatchTextReverser(), concurrency=2, parallel=True)
        .append("reverser1", BatchTextReverser(), concurrency=1, parallel=True)
        .append("reverser2", BatchTextReverser(), concurrency=0)
        .append("duplicator", BatchTextDuplicator(), concurrency=2, parallel=True)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 11, pipeline)


def test_mixed_concurrent_run():
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(96))
        .append("reverser0", TextReverser(), concurrency=2)
        .append("reverser1", BatchTextReverser(), concurrency=0)
        .append("reverser2", TextReverser(), concurrency=1)
        .append("duplicator", BatchTextDuplicator(), concurrency=2)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 96, pipeline)
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(81))
        .append("reverser0", BatchTextReverser(), concurrency=2, parallel=True)
        .append("reverser1", TextReverser(), concurrency=1, parallel=True)
        .append("reverser2", TextReverser(), concurrency=0)
        .append("duplicator", BatchTextDuplicator(), concurrency=2, parallel=True)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 81, pipeline)
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(92))
        .append("reverser0", BatchTextReverser(), concurrency=0)
        .append("reverser1", TextReverser(), concurrency=1)
        .append("duplicator", BatchTextDuplicator(), concurrency=0)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 92, pipeline)
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(83))
        .append("reverser0", BatchTextReverser(), concurrency=1, parallel=True)
        .append("reverser1", BatchTextReverser(), concurrency=1, parallel=False)
        .append("duplicator", TextDuplicator(), concurrency=1, parallel=True)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 83, pipeline)
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(100))
        .append("duplicator0", TextDuplicator(), concurrency=0)
        .append("reverser", TextReverser(), concurrency=0)
        .append("duplicator1", BatchTextDuplicator(check_batch=True), concurrency=0)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 100, pipeline)
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(100))
        .append("duplicator0", TextDuplicator(), concurrency=2)
        .append("reverser", TextReverser(), concurrency=2)
        .append("duplicator1", BatchTextDuplicator(), concurrency=2)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 100, pipeline)


def test_concurrency_errors():
    with pytest.raises(Exception):
        pipeline = (
            get_pipeline()
            .set_source(RandomTextSource(29))
            .append("reverser", BatchTextReverser(size=7), concurrency=1)
            .append("error", ExceptionStage(), concurrency=1)
            .build()
        )
        try:
            for _ in pipeline.run():
                pass
        except Exception:
            tb = traceback.format_exc()
            assert 'Exception("test exception")' in tb
            assert "CriticalError" in tb
            raise
        assert pipeline.count == 1
    with pytest.raises(Exception):
        pipeline = (
            get_pipeline()
            .set_source(RandomTextSource(10))
            .append("reverser", BatchTextReverser(size=3), concurrency=1, parallel=True)
            .append("error", ExceptionStage(), concurrency=1, parallel=True)
            .build()
        )
        for _ in pipeline.run():
            pass
        assert pipeline.count == 1
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(29))
        .append("reverser", BatchTextReverser(size=3), concurrency=3)
        .append("error", ErrorStage(), concurrency=1)
        .append("duplicator", BatchTextDuplicator(size=7), concurrency=2)
        .build()
    )
    for item in pipeline.run():
        assert item.get_timing("reverser")
        assert item.get_timing("duplicator")
        assert any(k.startswith("text_") for k in item.data.keys())
        assert item.get_timing("error")
    assert pipeline.count == 29
    with pytest.raises(Exception):
        pipeline = (
            get_pipeline()
            .set_source(RandomTextSource(29))
            .append("reverser", BatchTextReverser(size=7), concurrency=1)
            .append("error", BatchExceptionStage(size=3), concurrency=1)
            .build()
        )
        try:
            for _ in pipeline.run():
                pass
        except Exception:
            tb = traceback.format_exc()
            assert 'Exception("test exception")' in tb
            assert "CriticalError" in tb
            raise
        assert pipeline.count == 1
    with pytest.raises(Exception):
        pipeline = (
            get_pipeline()
            .set_source(RandomTextSource(10))
            .append("reverser", BatchTextReverser(size=3), concurrency=1, parallel=True)
            .append("error", BatchExceptionStage(size=1), concurrency=1, parallel=True)
            .build()
        )
        try:
            for _ in pipeline.run():
                pass
        except Exception:
            tb = traceback.format_exc()
            assert 'Exception("test exception")' in tb
            assert "CriticalError" in tb
            raise
        assert pipeline.count == 1
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(29))
        .append("reverser", BatchTextReverser(size=3), concurrency=3)
        .append("error", BatchErrorStage(size=13), concurrency=1)
        .append("duplicator", BatchTextDuplicator(size=7), concurrency=2)
        .build()
    )
    for item in pipeline.run():
        assert item.get_timing("reverser")
        assert item.get_timing("duplicator")
        assert any(k.startswith("text_") for k in item.data.keys())
        assert item.get_timing("error")
    assert pipeline.count == 29


def test_concurrent_initialization():
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(100))
        .append_concurrently(
            "reverser0", BatchTextReverser, kwargs={"cycles": 3}, concurrency=2
        )
        .append_concurrently("reverser1", BatchTextReverser, args=[5], concurrency=0)
        .append("reverser2", BatchTextReverser(), concurrency=1)
        .append("duplicator", BatchTextDuplicator(), concurrency=2)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 100, pipeline)
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(100))
        .append_concurrently(
            "reverser0", BatchTextReverser, concurrency=2, parallel=True
        )
        .append_concurrently(
            "reverser1", BatchTextReverser, args=[10], concurrency=1, parallel=True
        )
        .append_concurrently("reverser2", BatchTextReverser, concurrency=0)
        .append_concurrently(
            "duplicator",
            BatchTextDuplicator,
            args=[10],
            concurrency=2,
            parallel=True,
        )
        .build()
    )
    items = list(pipeline.run())
    _check(items, 100, pipeline)
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(100))
        .append("reverser0", BatchTextReverser(), concurrency=0)
        .append_concurrently("reverser1", BatchTextReverser, concurrency=1)
        .append("duplicator", BatchTextDuplicator(10), concurrency=0)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 100, pipeline)
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(100))
        .append_concurrently("duplicator0", BatchTextDuplicator, concurrency=0)
        .append("reverser", BatchTextReverser(), concurrency=0)
        .append_concurrently("duplicator1", BatchTextDuplicator, concurrency=0)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 100, pipeline)
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(100))
        .append("reverser", BatchTextReverser(12), concurrency=0)
        .append_concurrently("duplicator0", BatchTextDuplicator, concurrency=0)
        .append("duplicator1", BatchTextDuplicator(), concurrency=1)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 100, pipeline)
    pipeline = (
        get_pipeline(max_init_workers=1)
        .set_source(RandomTextSource(100))
        .append_concurrently(
            "reverser0", BatchTextReverser, args=[20], concurrency=1, parallel=True
        )
        .append_concurrently(
            "reverser1", BatchTextReverser, args=[20], concurrency=1, parallel=False
        )
        .append_concurrently(
            "duplicator",
            BatchTextDuplicator,
            args=[20],
            concurrency=1,
            parallel=True,
        )
        .build()
    )
    items = list(pipeline.run())
    _check(items, 100, pipeline)


def test_huge_run():
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(200))
        .append("reverser0", BatchTextReverser(10000), concurrency=2, parallel=True)
        .append("reverser1", BatchTextReverser(10000), concurrency=3, parallel=True)
        .append("reverser2", BatchTextReverser(10000), concurrency=1, parallel=True)
        .append("duplicator", BatchTextDuplicator(10000), concurrency=2, parallel=True)
        .build()
    )
    start_time = time.time()
    items = list(pipeline.run())
    elapsed1 = time.time() - start_time
    logger.debug("Time for parallel: %s", elapsed1)
    _check(items, 200)
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(200))
        .append("reverser0", BatchTextReverser(10000), concurrency=0)
        .append("reverser1", BatchTextReverser(10000), concurrency=0)
        .append("reverser2", BatchTextReverser(10000), concurrency=0)
        .append("duplicator", BatchTextDuplicator(10000), concurrency=0)
        .build()
    )
    start_time = time.time()
    items = list(pipeline.run())
    elapsed2 = time.time() - start_time
    logger.debug("Time for sequential: %s", elapsed2)
    _check(items, 200)
    assert elapsed2 > elapsed1


def test_timeouts():
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(100))
        .append("reverser0", BatchTextReverser(timeout=1, size=120), concurrency=0)
        .append("reverser1", BatchTextReverser(timeout=1, size=120), concurrency=0)
        .append("reverser2", BatchTextReverser(timeout=0, size=120), concurrency=0)
        .append("duplicator", BatchTextDuplicator(timeout=1, size=120), concurrency=0)
        .build()
    )
    start_time = time.time()
    items = list(pipeline.run())
    elapsed = time.time() - start_time
    assert 3 <= round(elapsed)
    _check(items, 100, pipeline)
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(100))
        .append("reverser0", BatchTextReverser(timeout=1, size=120), concurrency=1)
        .append("reverser1", BatchTextReverser(timeout=2, size=120), concurrency=1)
        .append("reverser2", BatchTextReverser(timeout=0, size=120), concurrency=1)
        .append("duplicator", BatchTextDuplicator(timeout=1, size=120), concurrency=1)
        .build()
    )
    start_time = time.time()
    items = list(pipeline.run())
    elapsed = time.time() - start_time
    assert 4 <= round(elapsed)
    _check(items, 100, pipeline)
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(100))
        .append(
            "reverser0",
            BatchTextReverser(timeout=1, size=120),
            concurrency=1,
            parallel=True,
        )
        .append(
            "reverser1",
            BatchTextReverser(timeout=0, size=120),
            concurrency=1,
            parallel=True,
        )
        .append(
            "reverser2",
            BatchTextReverser(timeout=1, size=120),
            concurrency=1,
            parallel=True,
        )
        .append(
            "duplicator",
            BatchTextDuplicator(timeout=0, size=120),
            concurrency=1,
            parallel=True,
        )
        .build()
    )
    start_time = time.time()
    items = list(pipeline.run())
    elapsed = time.time() - start_time
    assert 2 <= round(elapsed)
    _check(items, 100, pipeline)


def _check_item(item):
    assert item.data
    item.metadata["check"] = True


def test_single_items(items_generator_fx):
    pipeline = (
        get_pipeline()
        .append("reverser0", BatchTextReverser())
        .append("reverser1", BatchTextReverser())
        .append("reverser2", BatchTextReverser())
        .append("duplicator", BatchTextDuplicator())
        .build()
    )
    item = next(items_generator_fx)
    for _ in range(88):
        pipeline.process_async(copy.deepcopy(item), callback=_check_item)
    pipeline.stop()
    result = pipeline.get_item()
    assert result.id == item.id
    assert result.metadata["check"]
    assert result.data["text"] != item.data["text"]
    for _ in range(87):
        pipeline.get_item()

    pipeline = (
        get_pipeline()
        .append_concurrently(
            "reverser0", BatchTextReverser, kwargs={"cycles": 3}, concurrency=2
        )
        .append_concurrently("reverser1", BatchTextReverser, args=[5], concurrency=0)
        .append("reverser2", BatchTextReverser(), concurrency=1)
        .append("duplicator", BatchTextDuplicator(), concurrency=2)
        .build()
    )
    item = next(items_generator_fx)
    pipeline.process_async(copy.deepcopy(item), callback=_check_item)
    pipeline.stop()
    result = pipeline.get_item()
    assert result.id == item.id
    assert result.metadata["check"]
    assert result.data["text"] != item.data["text"]

    pipeline = (
        get_pipeline()
        .append_concurrently(
            "reverser0", BatchTextReverser, concurrency=2, parallel=True
        )
        .append_concurrently(
            "reverser1", BatchTextReverser, args=[9], concurrency=1, parallel=True
        )
        .append_concurrently("reverser2", BatchTextReverser, concurrency=0)
        .append_concurrently(
            "duplicator",
            BatchTextDuplicator,
            args=[10],
            concurrency=2,
            parallel=True,
        )
        .build()
    )
    item = next(items_generator_fx)
    for _ in range(88):
        pipeline.process_async(item, callback=_check_item)
    pipeline.stop()
    for _ in range(88):
        result = pipeline.get_item()
        assert result.id == item.id
        assert result.metadata["check"]
        assert result.data["text"]
        assert result.data["text"] != item.data["text"]

    pipeline = (
        get_pipeline()
        .append("reverser0", BatchTextReverser(), concurrency=0)
        .append_concurrently("reverser1", BatchTextReverser, concurrency=1)
        .append("duplicator", BatchTextDuplicator(10), concurrency=0)
        .build()
    )
    item = next(items_generator_fx)
    pipeline.process_async(copy.deepcopy(item), callback=_check_item)
    pipeline.stop()
    result = pipeline.get_item()
    assert result.id == item.id
    assert result.metadata["check"]
    assert result.data["text"] == item.data["text"]

    pipeline = (
        get_pipeline()
        .append_concurrently("duplicator0", BatchTextDuplicator)
        .append("reverser", BatchTextReverser())
        .append_concurrently("duplicator1", BatchTextDuplicator)
        .build()
    )
    item = next(items_generator_fx)
    pipeline.process_async(copy.deepcopy(item), callback=_check_item)
    pipeline.stop()
    result = pipeline.get_item()
    assert result.id == item.id
    assert result.metadata["check"]
    assert result.data["text"] != item.data["text"]

    pipeline = (
        get_pipeline()
        .append("reverser", BatchTextReverser(11), concurrency=0)
        .append_concurrently("duplicator0", BatchTextDuplicator, concurrency=0)
        .append("duplicator1", BatchTextDuplicator(), concurrency=1)
        .build()
    )
    item = next(items_generator_fx)
    pipeline.process_async(copy.deepcopy(item), callback=_check_item)
    pipeline.stop()
    result = pipeline.get_item()
    assert result.id == item.id
    assert result.metadata["check"]
    assert result.data["text"] != item.data["text"]

    pipeline = (
        get_pipeline(max_init_workers=2)
        .append_concurrently(
            "reverser0", BatchTextReverser, args=[20], concurrency=1, parallel=True
        )
        .append_concurrently(
            "reverser1", BatchTextReverser, args=[20], concurrency=1, parallel=False
        )
        .append_concurrently(
            "duplicator",
            BatchTextDuplicator,
            args=[20],
            concurrency=1,
            parallel=True,
        )
        .build()
    )
    item = next(items_generator_fx)
    pipeline.process_async(item, callback=_check_item)
    pipeline.stop()
    result = pipeline.get_item()
    assert result.id == item.id
    assert result.metadata["check"]
    assert result.data["text"] == item.data["text"]
    assert len(result.data.keys()) > len(item.data.keys())


def test_mixed_single_items(items_generator_fx):
    pipeline = (
        get_pipeline()
        .append("reverser0", BatchTextReverser())
        .append("reverser1", TextReverser())
        .append("reverser2", BatchTextReverser())
        .append("duplicator", TextDuplicator())
        .build()
    )
    item = next(items_generator_fx)
    for _ in range(88):
        pipeline.process_async(copy.deepcopy(item), callback=_check_item)
    pipeline.stop()
    result = pipeline.get_item()
    assert result.id == item.id
    assert result.metadata["check"]
    assert result.data["text"] != item.data["text"]
    for _ in range(87):
        pipeline.get_item()

    pipeline = (
        get_pipeline()
        .append_concurrently(
            "reverser0", TextReverser, kwargs={"cycles": 3}, concurrency=2
        )
        .append_concurrently("reverser1", BatchTextReverser, args=[5], concurrency=0)
        .append("reverser2", TextReverser(), concurrency=1)
        .append("duplicator", BatchTextDuplicator(), concurrency=2)
        .build()
    )
    item = next(items_generator_fx)
    pipeline.process_async(copy.deepcopy(item), callback=_check_item)
    pipeline.stop()
    result = pipeline.get_item()
    assert result.id == item.id
    assert result.metadata["check"]
    assert result.data["text"] != item.data["text"]

    pipeline = (
        get_pipeline()
        .append_concurrently("reverser0", TextReverser, concurrency=2, parallel=True)
        .append_concurrently(
            "reverser1", BatchTextReverser, args=[9], concurrency=1, parallel=True
        )
        .append_concurrently("reverser2", TextReverser, concurrency=0)
        .append_concurrently(
            "duplicator",
            BatchTextDuplicator,
            args=[10],
            concurrency=2,
            parallel=True,
        )
        .build()
    )
    item = next(items_generator_fx)
    for _ in range(88):
        pipeline.process_async(item, callback=_check_item)
    pipeline.stop()
    for _ in range(88):
        result = pipeline.get_item()
        assert result.id == item.id
        assert result.metadata["check"]
        assert result.data["text"]
        assert result.data["text"] != item.data["text"]



from __future__ import annotations

import logging
import time
import uuid
from collections import OrderedDict
from concurrent.futures import ProcessPoolExecutor
from concurrent.futures._base import Executor, Future
from concurrent.futures.thread import ThreadPoolExecutor
from inspect import isclass
from logging.handlers import QueueHandler
from multiprocessing import Manager, get_context
from multiprocessing.managers import SyncManager
from queue import Queue
from threading import Event, Thread
from typing import Any, Callable, Generator, Mapping, Optional
from typing import OrderedDict as OrderedDictType
from typing import Sequence, Tuple, Type, Union

from smartpipeline.containers import (
    BatchConcurrentStageContainer,
    BatchStageContainer,
    ConcurrentStageContainer,
    ConnectedStageMixin,
    ContainerType,
    SourceContainer,
    StageContainer,
)
from smartpipeline.defaults import CONCURRENCY_WAIT, MAX_QUEUES_SIZE
from smartpipeline.error.handling import ErrorManager, RetryManager
from smartpipeline.item import Item, Stop
from smartpipeline.stage import BatchStage, ItemsQueue, Source, Stage, StageType
from smartpipeline.utils import LogsReceiver, ProcessCounter, ThreadCounter, last_key

__author__ = "Giacomo Berardi <giacbrd.com>"


def _stage_initialization_with_logger(
    logs_queue: Queue, stage_class: Type[StageType], args: Sequence, kwargs: Mapping
) -> StageType:
    root_logger = logging.getLogger()
    # only by comparing string of queues we obtain their "original" address
    if not any(
        isinstance(handler, QueueHandler) and str(handler.queue) == str(logs_queue)
        for handler in root_logger.handlers
    ):
        handler = QueueHandler(logs_queue)
        root_logger.addHandler(handler)
    return stage_class(*args, **kwargs)


class EmptyStage(Stage):
    def process(self, item: Item) -> Item:
        return item


class Pipeline:
    def __init__(
        self,
        max_init_workers: Optional[int] = None,
        max_queues_size: int = MAX_QUEUES_SIZE,
    ):
        """
        :param max_init_workers: Number of workers to use for concurrent initialization of stages, default the number of CPUs
        :param max_queues_size: Maximum size of any queue instanced for the pipeline (stage input and output queues)
        """
        self._containers: OrderedDictType[str, ContainerType] = OrderedDict()
        self._error_manager = ErrorManager()
        self._max_init_workers = max_init_workers
        self._init_executor: Optional[Executor] = None
        self._wait_previous_executor: Optional[Executor] = None
        self._pipeline_executor: Optional[Thread] = None
        self._max_queues_size = max_queues_size
        self._out_queue = None
        self._enqueue_source = False
        # we approach sharing memory between processes exclusively through a `multiprocessing.Manager`
        self._sync_manager: Optional[SyncManager] = None
        self._count = 0
        self._executors_ready = False
        self._logs_receiver: Optional[LogsReceiver] = None
        self._name = f"{self.__class__.__name__}-{str(uuid.uuid4())[:8]}"
        self._logger = logging.getLogger(self._name)
        self._source_name = f"SourceOf{self._name}"
        # Initialize with a trivial "support" source, on which we can eventually send items when using `process` methods
        self._source_container = SourceContainer(self._source_name)
        # A fake container for the pipeline to support data structures initialization
        self._fake_container = StageContainer(
            f"FakeContainerOf{self._name}", EmptyStage(), ErrorManager(), RetryManager()
        )

    def _start_logs_receiver(self) -> LogsReceiver:
        if self._logs_receiver is None:
            if self._sync_manager is None:
                self._sync_manager = Manager()
            logs_queue = self._sync_manager.Queue()
            self._logs_receiver = LogsReceiver(logs_queue)
            self._logs_receiver.start()
        return self._logs_receiver

    def _stop_logs_receiver(self):
        if self._logs_receiver is not None:
            self._logs_receiver.stop()
            self._logs_receiver = None

    def _get_logs_receiver_queue(self) -> Queue:
        if self._logs_receiver is not None:
            return self._logs_receiver.queue
        else:
            raise AttributeError("Logs queue for multiprocessing never initialized")

    @property
    def name(self) -> str:
        """
        Pipeline unique name, equivalent to its logger name
        """
        return self._name

    def _new_mp_queue(self) -> ItemsQueue:
        """
        Construct queue for multiprocessing communication
        """
        # why not use multiprocessing.Queue or JoinableQueue instead of the manager?
        # answer https://stackoverflow.com/a/45236748 and the lack of "safety" of task_done() and join()
        if self._sync_manager is None:
            self._sync_manager = Manager()
        return self._sync_manager.Queue(maxsize=self._max_queues_size)

    def _new_queue(self) -> ItemsQueue:
        """
        Construct queue for communication
        """
        return Queue(maxsize=self._max_queues_size)

    def _new_mp_event(self) -> Event:
        """
        Construct synchronization event for multiprocessing
        """
        if self._sync_manager is None:
            self._sync_manager = Manager()
        return self._sync_manager.Event()

    @staticmethod
    def _new_event() -> Event:
        """
        Construct synchronization event
        """
        return Event()

    def _new_mp_counter(self) -> ProcessCounter:
        """
        Construct a safe counter for multiprocessing
        """
        if self._sync_manager is None:
            self._sync_manager = Manager()
        return ProcessCounter(self._sync_manager)

    @staticmethod
    def _new_counter() -> ThreadCounter:
        """
        Construct a safe counter for threads
        """
        return ThreadCounter()

    def _wait_executors(self, wait_seconds: float = CONCURRENCY_WAIT):
        """
        Wait for all containers to start

        :param wait_seconds: Recurrently wait these seconds for all stage initializers to finish
        """
        if self._executors_ready:
            return
        if self._init_executor is not None:
            self._init_executor.shutdown(wait=True)
            self._init_executor = None
        while not all(self._containers.values()):
            time.sleep(wait_seconds)
        if self._wait_previous_executor is not None:
            self._wait_previous_executor.shutdown(wait=True)
        for name, container in self._containers.items():
            if isinstance(
                container, (ConcurrentStageContainer, BatchConcurrentStageContainer)
            ):
                container.run()
        # finalize initialization of the error manager shared by this and other stage threads
        self._error_manager.on_start()
        self._executors_ready = True
        self._logger.debug("Pipeline ready to run")

    def shutdown(self):
        """
        Execute shutdown of various pool executors and multiprocessing stuff.
        The developer should not need to call it explicitly.
        """
        if self._out_queue is not None:
            self._out_queue.join()
        if self._init_executor is not None:
            self._init_executor.shutdown()
        for name, container in self._containers.items():
            if isinstance(
                container, (ConcurrentStageContainer, BatchConcurrentStageContainer)
            ):
                container.shutdown()
        if self._sync_manager is not None:
            self._sync_manager.shutdown()

    def build(self) -> Pipeline:
        """
        Pipeline builder method
        """
        if not any(self._containers):
            raise ValueError("Must append at least a stage")
        self._logger.debug("Building the pipeline on stages: %s", self._log_stages())
        self._wait_executors()
        return self

    def _exit_run(self, source_thread, terminator_thread):
        self.stop()
        if source_thread is not None:
            source_thread.join()
        # TODO in case of errors we loose pending items!
        self._terminate_all(force=True)
        if terminator_thread is not None:
            terminator_thread.join()
        self.shutdown()
        self._count += 1

    def run(self) -> Generator[Item, None, None]:
        """
        Run the pipeline given a source and a concatenation of stages.
        Get the sequence of items through iteration

        :return: Iterator over processed items
        :raises ValueError: When a source has not been set for the pipeline
        """
        if not self._source_container.is_set():
            raise ValueError("Set the data source for this pipeline")
        self._logger.debug("Running the pipeline on stages: %s", self._log_stages())
        counter = 0
        last_stage_name = self._last_stage_name()
        terminator_thread = None
        source_errors = None
        source_thread = None
        # in case the first stage is concurrent
        if self._enqueue_source:
            source_errors = Queue()
            source_thread = Thread(
                target=self._source_container.pop_into_queue,
                kwargs={"errors_queue": source_errors},
            )
            source_thread.start()
        while True:
            for name, container in self._containers.items():
                try:
                    # first, check errors from the source
                    if source_errors is not None and not source_errors.empty():
                        source_error = source_errors.get(block=False)
                        if source_error is not None:
                            raise source_error
                    # concurrent stages run by themselves in threads/processes
                    if not isinstance(
                        container,
                        (ConcurrentStageContainer, BatchConcurrentStageContainer),
                    ):
                        container.process()
                    # but me must periodically check for errors
                    else:
                        container.check_errors()
                except Exception as e:
                    self._exit_run(source_thread, terminator_thread)
                    raise e
                # retrieve finally processed items from the last stage
                if name == last_stage_name:
                    for _ in range(
                        container.size
                        if isinstance(container, BatchStageContainer)
                        else 1
                    ):
                        item = container.get_processed()
                        if item is not None:
                            if not isinstance(item, Stop):
                                try:
                                    yield item
                                except GeneratorExit:
                                    # when the caller exits the run loop terminates everything
                                    self._exit_run(source_thread, terminator_thread)
                                    return
                                finally:
                                    counter += 1
                                self._count += 1
                            # if a stop is finally signaled, start termination of all containers
                            elif (
                                not self._all_terminated() and terminator_thread is None
                            ):
                                terminator_thread = Thread(target=self._terminate_all)
                                terminator_thread.start()
                        # an item is None if the final output queue is empty
                        else:
                            break
            # exit the loop only when all items have been returned
            if self._all_empty() and counter >= self._source_container.count():
                if source_thread is not None:
                    source_thread.join()
                if terminator_thread is not None:
                    terminator_thread.join()
                    self.shutdown()
                return

    @property
    def count(self) -> int:
        """
        Get the number of items processed by all executed runs, also for items which have failed

        :return: Count of processed items
        """
        return self._count

    def _terminate_all(
        self, force: bool = False, wait_seconds: float = CONCURRENCY_WAIT
    ):
        """
        Terminate all running containers

        :param force: If True do not wait for a container to process all items produced by the source
        :param wait_seconds: Time to wait before pinging again a container for its termination
        """
        self._logger.debug("Terminating the pipeline")
        # scroll the pipeline by its order (this is critical for correctness)
        # and terminate stages after the relative queues are empty
        for container in self._containers.values():
            if not force:
                # ensure the stage has processed all source items
                while container.count() < self._source_container.count():
                    time.sleep(wait_seconds)
            container.terminate()
            if isinstance(container, ConcurrentStageContainer):
                if force:
                    # empty the queues, losing pending items
                    container.empty_queues()
                while not container.is_terminated():
                    time.sleep(wait_seconds)
                container.queues_join()
                while not container.queues_empty():
                    time.sleep(wait_seconds)
            if (
                not isinstance(container, ConcurrentStageContainer)
                or not container.parallel
            ):
                container.stage.on_end()
        self._error_manager.on_end()
        self._stop_logs_receiver()
        self._logger.debug("Termination done")

    def _all_terminated(self) -> bool:
        """
        Check if all containers have been alerted for termination and are exited
        """
        return all(container.is_terminated() for container in self._containers.values())

    def _all_empty(self) -> bool:
        """
        Check if all containers are terminated and there are not items left in the queues
        """
        return self._all_terminated() and all(
            container.queues_empty()
            for container in self._containers.values()
            if isinstance(
                container, (ConcurrentStageContainer, BatchConcurrentStageContainer)
            )
        )

    def process(self, item: Item) -> Item:
        """
        Process a single item synchronously (no concurrency) through the pipeline
        """
        self._logger.debug("Processing %s on stages: %s", item, self._log_stages())
        last_stage_name = last_key(self._containers)
        self._source_container.prepend_item(item)
        ret = None
        for name, container in self._containers.items():
            container.process()
            if name == last_stage_name:
                ret = container.get_processed(block=True)
        return ret or item

    def process_async(
        self, item: Item, callback: Optional[Callable[[Item], Any]] = None
    ):
        """
        Process a single item asynchronously through the pipeline, stages may run concurrently.
        The call returns immediately, processed items are retrieved with :meth:`.Pipeline.get_item`

        :param callback: A function to call after a successful process of the item
        """
        self._logger.debug(
            "Processing asynchronously %s on stages: %s", item, self._log_stages()
        )
        if callback is not None:
            item.set_callback(callback)
        self._source_container.prepend_item(item)
        self._start_pipeline_executor()

    def stop(self):
        """
        Tell the source to stop to generate items and consequently terminate the pipeline
        after all "remaining" items are processed.
        """
        self._source_container.stop()

    def get_item(self, block: bool = True) -> Item:
        """
        Get a single item from the asynchronous execution of the pipeline on single items from :meth:`.Pipeline.process_async`

        :param block: If True wait indefinitely for the next processed item
        :raises ValueError: When there is not output queue set, the pipeline is not running asynchronously
        :raises queue.Empty: When we do not block and the queue is empty
        """
        if self._out_queue is not None:
            item = self._out_queue.get(block)
            self._out_queue.task_done()
            return item
        else:
            raise ValueError(
                "No pipeline is running asynchronously, not item can be retrieved from the output queue"
            )

    def set_source(self, source: Source) -> Pipeline:
        """
        Set the source of the pipeline: a subclass of :class:`.stage.Source`
        """
        source.set_name(self._source_name)
        self._source_container.set(source)
        return self

    def set_error_manager(self, error_manager: ErrorManager) -> Pipeline:
        """
        Set the error manager for handling errors from each stage item processing
        """
        self._error_manager = error_manager
        for container in self._containers.values():
            container.set_error_manager(self._error_manager)
        return self

    def _last_stage_name(self) -> Optional[str]:
        if self._containers:
            return last_key(self._containers)
        return None

    def _wait_for_previous(
        self,
        container: ConnectedStageMixin,
        last_stage_name: Optional[str],
        wait_seconds: float = CONCURRENCY_WAIT,
    ):
        """
        Given a container we want to append to the pipeline, wait for the last one (added to the pipeline) to be created

        :param container: A container to add to the pipeline
        :param last_stage_name: Name of the last stage currently in the pipeline
        :param wait_seconds: Time to recurrently wait the construction of the container relative to the last stage in the pipeline
        """

        def _waiter():
            if last_stage_name is not None:
                while self._containers[last_stage_name] is self._fake_container:
                    time.sleep(wait_seconds)
                container.set_previous(self._containers[last_stage_name])
            else:
                container.set_previous(self._source_container)

        executor = self._get_wait_previous_executor()
        executor.submit(_waiter)

    def _build_container(
        self,
        name: str,
        stage: StageType,
        concurrency: int,
        parallel: bool,
        retryable_errors: Tuple[Type[Exception], ...],
        max_retries: int,
        backoff: Union[float, int],
    ) -> ContainerType:
        """
        Get a new container instance according to the pipeline configuration

        :param name: Stage name
        :param stage: A stage instance
        :param concurrency: Number of concurrent stage executions, if 0 then just create the non-concurrent containers
        :param parallel: If True use multiprocessing, otherwise threads
        :param retryable_errors: list of exceptions for which the stage applies an exponential backoff strategy. When the maximum number of retries is hit, a :class:`.error.exceptions.RetryError` is raised
        :param max_retries: maximum number of retries for the stage before raising a RetryError(SoftError) (default 0)
        :param backoff: backoff factor for the `exponential backoff strategy` used by the stage when it raises one of the exceptions declared in `retryable_errors` param (default 0.0)
        """
        retry_manager = RetryManager(
            backoff=backoff, max_retries=max_retries, retryable_errors=retryable_errors
        )
        if concurrency <= 0:
            constructor = (
                BatchStageContainer if isinstance(stage, BatchStage) else StageContainer
            )
            # if not concurrent we must explicitly finalize initialization of this single stage object
            stage.on_start()
            return constructor(name, stage, self._error_manager, retry_manager)
        else:
            constructor = (
                BatchConcurrentStageContainer
                if isinstance(stage, BatchStage)
                else ConcurrentStageContainer
            )
            if parallel:
                self._start_logs_receiver()
                logs_queue = self._get_logs_receiver_queue()
                return constructor(
                    name,
                    stage,
                    self._error_manager,
                    retry_manager,
                    self._new_mp_queue,
                    self._new_mp_counter,
                    self._new_mp_event,
                    concurrency,
                    parallel,
                    logs_queue,
                )
            else:
                # if the stage is executed on multiple threads we must finalize initialization once,
                # while on multiprocessing each process executor calls it for its own copy of the stage
                stage.on_start()
                return constructor(
                    name,
                    stage,
                    self._error_manager,
                    retry_manager,
                    self._new_queue,
                    self._new_counter,
                    self._new_event,
                    concurrency,
                    parallel,
                )

    def get_stage(self, name: str) -> Optional[StageType]:
        """
        Get a stage instance by its name
        """
        ret = self._containers.get(name)
        return ret.stage if ret else None

    def append(
        self,
        name: str,
        stage: StageType,
        concurrency: int = 0,
        parallel: bool = False,
        retryable_errors: Tuple[Type[Exception], ...] = tuple(),
        max_retries: int = 0,
        backoff: Union[float, int] = 0.0,
    ) -> Pipeline:
        """
        Append a stage to the pipeline just after the last one appended, or after the source if it is the first stage

        :param name: Name for identify the stage in the pipeline, it is also set in the stage and it must be unique in the pipeline
        :param stage: Instance of a stage
        :param concurrency: Number of concurrent stage executions, if 0 then threads/processes won't be involved for this stage
        :param parallel: If True use multiprocessing, otherwise threads
        :param retryable_errors: list of exceptions for which the stage applies an exponential backoff strategy. When the maximum number of retries is hit, a :class:`.error.exceptions.RetryError` is raised
        :param max_retries: maximum number of retries for the stage before raising a RetryError(SoftError) (default 0)
        :param backoff: backoff factor for the `exponential backoff strategy` used by the stage when it raises one of the exceptions declared in `retryable_errors` param (default 0.0)
        """
        self._executors_ready = False
        # FIXME here we force a BatchStage to run on a thread, but we would leave it on the main thread
        if concurrency < 1 and isinstance(stage, BatchStage):
            parallel = False
            concurrency = 1
        self._check_stage_name(name)
        self._check_retries_params(retryable_errors, max_retries, backoff)
        container = self._build_container(
            name, stage, concurrency, parallel, retryable_errors, max_retries, backoff
        )
        if concurrency > 0:
            # if it is concurrent and it is the first stage, make the source work on a output queue
            if not self._containers:
                self._enqueue_source = True
        self._wait_for_previous(
            container, self._last_stage_name()
        )  # wait that previous stage is initialized
        self._containers[name] = container
        return self

    # deprecated
    append_stage = append  # pragma: no cover

    def append_concurrently(
        self,
        name: str,
        stage_class: Type[StageType],
        args: Optional[Sequence] = None,
        kwargs: Optional[Mapping] = None,
        concurrency: int = 0,
        parallel: bool = False,
        retryable_errors: Tuple[Type[Exception], ...] = tuple(),
        max_retries: int = 0,
        backoff: Union[float, int] = 0.0,
    ) -> Pipeline:
        """
        Append a stage class to the pipeline just after the last one appended, or after the source if it is the first stage.
        The stage construction will be executed concurrently respect to the general pipeline construction

        :param name: Name for identify the stage in the pipeline, it is also set in the stage and it must be unique in the pipeline
        :param stage_class: Class of a stage
        :param args: List of arguments for the stage constructor
        :param kwargs: Dictionary of keyed arguments for the stage constructor
        :param concurrency: Number of concurrent stage executions, if 0 then threads/processes won't be involved for this stage
        :param parallel: If True use multiprocessing, otherwise threads (also for concurrent construction)
        :param retryable_errors: list of exceptions for which the stage applies an exponential backoff strategy. When the maximum number of retries is hit, a :class:`.error.exceptions.RetryError` is raised
        :param max_retries: maximum number of retries for the stage before raising a RetryError(SoftError) (default 0)
        :param backoff: backoff factor for the `exponential backoff strategy` used by the stage when it raises one of the exceptions declared in `retryable_errors` param (default 0.0)
        """
        if not isclass(stage_class):
            raise ValueError(
                "Must pass stage classes, and not stage instances, as argument"
            )
        self._executors_ready = False
        parallel_construction = parallel
        # FIXME here we force a BatchStage to run on a thread, but we would like to leave it on the main thread
        if concurrency < 1 and issubclass(stage_class, BatchStage):
            parallel = False
            concurrency = 1
        if kwargs is None:
            kwargs = {}
        if args is None:
            args = []
        self._check_stage_name(name)
        self._check_retries_params(retryable_errors, max_retries, backoff)
        # if it is concurrent, and it is the first stage, make the source work on a output queue
        if concurrency > 0 and not self._containers:
            self._enqueue_source = True
        last_stage_name = self._last_stage_name()
        # set it immediately with an empty stage container so the order of the calls of this method is followed in `_containers`
        self._containers[name] = self._fake_container
        self._start_logs_receiver()
        logs_queue = self._get_logs_receiver_queue()
        executor = self._get_init_executor(parallel_construction)
        if parallel_construction:
            future = executor.submit(
                _stage_initialization_with_logger, logs_queue, stage_class, args, kwargs
            )
        else:
            future = executor.submit(stage_class, *args, **kwargs)

        def append_stage(stage_future: Future):
            stage = stage_future.result()
            container = self._build_container(
                name,
                stage,
                concurrency,
                parallel,
                retryable_errors,
                max_retries,
                backoff,
            )
            self._wait_for_previous(container, last_stage_name)
            self._containers[name] = container

        future.add_done_callback(append_stage)
        return self

    # deprecated
    append_stage_concurrently = append_concurrently

    def _get_init_executor(self, parallel: bool = False) -> Executor:
        """
        Get a pool executor for concurrent stage initialization

        :param parallel: True if the executor uses multiprocessing, otherwise treads
        """
        if self._init_executor is None:
            if not parallel:
                self._init_executor = ThreadPoolExecutor(
                    max_workers=self._max_init_workers
                )
            else:
                self._init_executor = ProcessPoolExecutor(
                    max_workers=self._max_init_workers, mp_context=get_context("spawn")
                )
        return self._init_executor

    def _get_wait_previous_executor(self) -> Executor:
        """
        Get a pool executor for the function that will recurrently wait for a container to be ready
        """
        if self._wait_previous_executor is None:
            self._wait_previous_executor = ThreadPoolExecutor()
        return self._wait_previous_executor

    def _start_pipeline_executor(self) -> Thread:
        """
        Get a thread where to run a pipeline that accepts asynchronous processing of single items
        """
        if self._pipeline_executor is None:
            self._init_out_queue()

            def pipeline_runner():
                for item in self.run():
                    item.callback()
                    self._out_queue.put(item)

            self._pipeline_executor = Thread(target=pipeline_runner)
            self._pipeline_executor.start()
        return self._pipeline_executor

    def _check_stage_name(self, name: str):
        """
        Check if a stage name is not already defined in the pipeline
        :raises ValueError: Stage name is already defined in the pipeline
        """
        if name in self._containers:
            raise ValueError(f"The stage name {name} is already used in this pipeline")

    @staticmethod
    def _check_retries_params(
        retryable_errors: Tuple[Type[Exception], ...],
        max_retries: int,
        backoff: Union[float, int],
    ):
        """
        Check the values used to apply the retry strategy for a stage
        """
        if not (isinstance(backoff, (float, int)) and backoff >= 0):
            raise ValueError(
                "The `backoff` parameter must be either a float or int and its value must be >= 0"
            )
        if not (isinstance(max_retries, int) and max_retries >= 0):
            raise ValueError(
                "The `max_retries` parameter must be an int and its value must be >= 0"
            )
        if not (
            isinstance(retryable_errors, tuple)
            and all(issubclass(err, Exception) for err in retryable_errors)
        ):
            raise ValueError(
                "The `retryable_errors` parameter must be a tuple and all its elements must be Exceptions"
            )

    def _init_out_queue(self):
        """
        Get the internal output pipeline for asynchronous processing of single items
        """
        self._out_queue = self._new_queue()

    def _log_stages(self):
        return ", ".join(self._containers.keys())



import copy
import logging
import queue
import time
import traceback
from logging.handlers import QueueHandler

import pytest

from smartpipeline.error.exceptions import RetryError
from smartpipeline.error.handling import ErrorManager
from smartpipeline.pipeline import Pipeline, _stage_initialization_with_logger
from tests.utils import (
    CriticalIOErrorStage,
    CustomizableBrokenStage,
    ErrorSerializableStage,
    ErrorStage,
    ExceptionStage,
    Logger,
    RandomTextSource,
    SerializableErrorManager,
    SerializableStage,
    SlowInitializer,
    TextDuplicator,
    TextReverser,
    TimeWaster,
    get_pipeline,
)

__author__ = "Giacomo Berardi <giacbrd.com>"

logger = logging.getLogger(__name__)


def _check(items, num, pipeline=None):
    diff = frozenset(range(1, num + 1)).difference(item.data["count"] for item in items)
    assert not diff, "Not found items: {}".format(", ".join(str(x) for x in diff))
    assert len(items) == num
    if pipeline:
        assert num == pipeline.count


def test_concurrent_run():
    # no stages
    with pytest.raises(ValueError):
        get_pipeline().build()
    # no source
    with pytest.raises(ValueError):
        pipeline = get_pipeline().append("reverser", TextReverser()).build()
        list(pipeline.run())
    pipeline = get_pipeline().append("reverser", TextReverser()).build()
    assert pipeline._new_mp_queue() != pipeline._new_mp_queue()
    assert pipeline._new_queue() != pipeline._new_queue()
    assert not isinstance(pipeline._new_queue(), pipeline._new_mp_queue().__class__)
    pipeline = get_pipeline().append("reverser", TextReverser()).build()
    assert pipeline._new_mp_counter() != pipeline._new_mp_counter()
    assert pipeline._new_counter() != pipeline._new_counter()
    assert not isinstance(pipeline._new_counter(), pipeline._new_mp_counter().__class__)
    pipeline = get_pipeline().append("reverser", TextReverser()).build()
    assert pipeline._new_mp_event() != pipeline._new_mp_event()
    assert pipeline._new_event() != pipeline._new_event()
    assert not isinstance(pipeline._new_event(), pipeline._new_mp_event().__class__)
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(100))
        .append("reverser0", TextReverser(), concurrency=2)
        .append("reverser1", TextReverser(), concurrency=0)
        .append("reverser2", TextReverser(), concurrency=1)
        .append("duplicator", TextDuplicator(), concurrency=2)
        .build()
    )
    items = list(pipeline.run())
    assert pipeline._wait_executors() is None
    _check(items, 100, pipeline)
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(100))
        .append("reverser0", TextReverser(), concurrency=2, parallel=True)
        .append("reverser1", TextReverser(), concurrency=1, parallel=True)
        .append("reverser2", TextReverser(), concurrency=0)
        .append("duplicator", TextDuplicator(), concurrency=2, parallel=True)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 100, pipeline)
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(100))
        .append("reverser0", TextReverser(), concurrency=0)
        .append("reverser1", TextReverser(), concurrency=1)
        .append("duplicator", TextDuplicator(), concurrency=0)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 100, pipeline)
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(100))
        .append("reverser0", TextReverser(), concurrency=1, parallel=True)
        .append("reverser1", TextReverser(), concurrency=1, parallel=False)
        .append("duplicator", TextDuplicator(), concurrency=1, parallel=True)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 100, pipeline)
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(100))
        .append("duplicator0", TextDuplicator(), concurrency=0)
        .append("reverser", TextReverser(), concurrency=0)
        .append("duplicator1", TextDuplicator(), concurrency=0)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 100, pipeline)
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(100))
        .append("reverser", TextReverser(), concurrency=0)
        .append("duplicator0", TextDuplicator(), concurrency=0)
        .append("duplicator1", TextDuplicator(), concurrency=1)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 100, pipeline)


def test_queue_sizes():
    pipeline = (
        get_pipeline(max_queues_size=1)
        .set_source(RandomTextSource(100))
        .append("reverser0", TextReverser(), concurrency=2)
        .append("reverser1", TextReverser(), concurrency=0)
        .append("reverser2", TextReverser(), concurrency=1)
        .append("duplicator", TextDuplicator(), concurrency=2)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 100, pipeline)
    pipeline = (
        get_pipeline(max_queues_size=1)
        .set_source(RandomTextSource(100))
        .append("reverser0", TextReverser(), concurrency=2, parallel=True)
        .append("reverser1", TextReverser(), concurrency=1, parallel=True)
        .append("reverser2", TextReverser(), concurrency=0)
        .append("duplicator", TextDuplicator(), concurrency=2, parallel=True)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 100, pipeline)
    pipeline = (
        get_pipeline(max_queues_size=3)
        .set_source(RandomTextSource(100))
        .append("waster", TimeWaster(0.02), concurrency=2)
        .append("reverser", TimeWaster(0.04), concurrency=1, parallel=True)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 100, pipeline)
    pipeline = (
        get_pipeline(max_queues_size=0)
        .set_source(RandomTextSource(100))
        .append("reverser0", TextReverser(), concurrency=2)
        .append("reverser1", TextReverser(), concurrency=0)
        .append("reverser2", TextReverser(), concurrency=1)
        .append("duplicator", TextDuplicator(), concurrency=2)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 100, pipeline)
    pipeline = (
        get_pipeline(max_queues_size=0)
        .set_source(RandomTextSource(100))
        .append("reverser0", TextReverser(), concurrency=2, parallel=True)
        .append("reverser1", TextReverser(), concurrency=1, parallel=True)
        .append("reverser2", TextReverser(), concurrency=0)
        .append("duplicator", TextDuplicator(), concurrency=2, parallel=True)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 100, pipeline)


def test_concurrency_errors(caplog):
    with pytest.raises(Exception):
        pipeline = (
            get_pipeline()
            .set_source(RandomTextSource(10))
            .append("reverser", TextReverser(), concurrency=1)
            .append("error", ExceptionStage(), concurrency=1)
            .build()
        )
        try:
            for _ in pipeline.run():
                pass
        except Exception:
            tb = traceback.format_exc()
            assert 'Exception("test exception")' in tb
            assert "CriticalError" in tb
            raise
        assert (
            len(
                [
                    record
                    for record in caplog.records
                    if record.levelno == logging.ERROR
                    and "has generated an error" in record.msg.lower()
                ]
            )
            == 1
        )
        assert pipeline.count == 1
    caplog.clear()
    with pytest.raises(Exception):
        pipeline = (
            get_pipeline()
            .set_source(RandomTextSource(10))
            .append("reverser", TextReverser(), concurrency=1, parallel=True)
            .append("error", ExceptionStage(), concurrency=1, parallel=True)
            .build()
        )
        try:
            for _ in pipeline.run():
                pass
        except Exception:
            tb = traceback.format_exc()
            assert 'Exception("test exception")' in tb
            assert "CriticalError" in tb
            raise
        assert (
            len(
                [
                    record
                    for record in caplog.records
                    if record.levelno == logging.ERROR
                    and "has generated an error" in record.msg.lower()
                ]
            )
            == 1
        )
        assert pipeline.count == 1
    caplog.clear()
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(10))
        .append("reverser", TextReverser(), concurrency=1)
        .append("error1", ErrorStage(), concurrency=1, parallel=True)
        .append("error2", ErrorStage(), concurrency=2)
        .append("duplicator", TextDuplicator(), concurrency=1, parallel=True)
        .build()
    )
    for item in pipeline.run():
        assert item.get_timing("reverser")
        assert item.get_timing("duplicator")
        assert any(k.startswith("text_") for k in item.data.keys())
        assert len(list(item.soft_errors())) == 2
        assert item.get_timing("error1")
        assert item.get_timing("error2")
    assert (
        len(
            [
                record
                for record in caplog.records
                if record.levelno == logging.ERROR
                and "has generated an error" in record.msg.lower()
            ]
        )
        == 20
    )
    assert pipeline.count == 10


def test_concurrent_constructions():
    """test `on_start` and `on_end` methods"""
    serializable_stage = SerializableStage()
    error_manager = SerializableErrorManager()
    pipeline = (
        get_pipeline()
        .set_error_manager(error_manager)
        .set_source(RandomTextSource(10))
        .append("reverser1", TextReverser(), concurrency=1, parallel=True)
        .append("stage", serializable_stage, concurrency=2, parallel=True)
        .append("reverser2", TextReverser())
        .append("error", ErrorStage(), concurrency=2, parallel=True)
        .build()
    )
    for item in pipeline.run():
        assert serializable_stage._file is None
        assert item.data.get("file")
        assert item.get_timing("reverser1")
        assert item.get_timing("reverser2")
        assert item.has_soft_errors()
    assert pipeline._wait_executors() is None
    assert pipeline.count == 10
    assert serializable_stage._file is None
    assert error_manager.is_closed()
    serializable_stage = SerializableStage()
    error_manager = SerializableErrorManager()
    pipeline = (
        get_pipeline()
        .set_error_manager(error_manager)
        .set_source(RandomTextSource(10))
        .append("stage", serializable_stage)
        .append("reverser1", TextReverser(), concurrency=1, parallel=True)
        .append("reverser2", TextReverser(), concurrency=1, parallel=True)
        .append("error", ErrorStage())
        .build()
    )
    for item in pipeline.run():
        assert item.data.get("file")
        assert item.get_timing("reverser1")
        assert item.get_timing("reverser2")
        assert item.has_soft_errors()
    assert pipeline.count == 10
    assert serializable_stage.is_closed()
    assert error_manager.is_closed()
    with pytest.raises(IOError):
        try:
            pipeline = get_pipeline()
            (
                pipeline.set_error_manager(SerializableErrorManager())
                .set_source(RandomTextSource(10))
                .append("stage", ErrorSerializableStage(), concurrency=1, parallel=True)
                .build()
            )
        finally:
            pipeline._stop_logs_receiver()
    error_manager = SerializableErrorManager().raise_on_critical_error()
    with pytest.raises(IOError):
        pipeline = (
            get_pipeline()
            .set_error_manager(error_manager)
            .set_source(RandomTextSource(10))
            .append("stage", CriticalIOErrorStage(), concurrency=1, parallel=True)
            .build()
        )
        for item in pipeline.run():
            pass
    assert error_manager.is_closed()


def test_concurrent_initialization():
    with pytest.raises(ValueError):
        get_pipeline().append_concurrently(
            "reverser0", TextReverser(), kwargs={"cycles": 3}, concurrency=2
        )
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(100))
        .append_concurrently(
            "reverser0", TextReverser, kwargs={"cycles": 3}, concurrency=2
        )
        .append_concurrently("reverser1", TextReverser, args=[5], concurrency=0)
        .append("reverser2", TextReverser(), concurrency=1)
        .append("duplicator", TextDuplicator(), concurrency=2)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 100, pipeline)
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(100))
        .append_concurrently("reverser0", TextReverser, concurrency=2, parallel=True)
        .append_concurrently(
            "reverser1", TextReverser, args=[10], concurrency=1, parallel=True
        )
        .append_concurrently("reverser2", TextReverser, concurrency=0)
        .append_concurrently(
            "duplicator", TextDuplicator, args=[10], concurrency=2, parallel=True
        )
        .build()
    )
    items = list(pipeline.run())
    _check(items, 100, pipeline)
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(100))
        .append("reverser0", TextReverser(), concurrency=0)
        .append_concurrently("reverser1", TextReverser, concurrency=1)
        .append("duplicator", TextDuplicator(10), concurrency=0)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 100, pipeline)
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(100))
        .append_concurrently("duplicator0", TextDuplicator, concurrency=0)
        .append("reverser", TextReverser(), concurrency=0)
        .append_concurrently("duplicator1", TextDuplicator, concurrency=0)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 100, pipeline)
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(100))
        .append("reverser", TextReverser(12), concurrency=0)
        .append_concurrently("duplicator0", TextDuplicator, concurrency=0)
        .append("duplicator1", TextDuplicator(), concurrency=1)
        .build()
    )
    items = list(pipeline.run())
    _check(items, 100, pipeline)
    pipeline = (
        get_pipeline(max_init_workers=1)
        .set_source(RandomTextSource(100))
        .append_concurrently(
            "reverser0", TextReverser, args=[20], concurrency=1, parallel=True
        )
        .append_concurrently(
            "reverser1", TextReverser, args=[20], concurrency=1, parallel=False
        )
        .append_concurrently(
            "duplicator", TextDuplicator, args=[20], concurrency=1, parallel=True
        )
        .build()
    )
    items = list(pipeline.run())
    _check(items, 100, pipeline)
    pipeline = (
        get_pipeline(max_init_workers=1)
        .set_source(RandomTextSource(100))
        .append_concurrently("slow_init", SlowInitializer, concurrency=1, parallel=True)
        .build()
    )
    pipeline._terminate_all()
    pipeline.shutdown()
    assert pipeline._containers["slow_init"]


def test_build():
    """Test the build pattern with non-conventional order of the method calls"""
    pipeline1 = (
        Pipeline()
        .append("reverser", TextReverser(), concurrency=2)
        .append("duplicator", TextDuplicator())
        .set_source(RandomTextSource(10))
        .build()
    )
    pipeline2 = (
        Pipeline()
        .append("reverser", TextReverser())
        .append_concurrently("duplicator", TextDuplicator, concurrency=2)
        .set_source(RandomTextSource(10))
        .build()
    )
    pipeline3 = (
        Pipeline()
        .append_concurrently("reverser", TextReverser, concurrency=1, parallel=True)
        .append_concurrently("duplicator", TextDuplicator, concurrency=2)
        .set_source(RandomTextSource(10))
        .build()
    )
    for pipeline in (pipeline1, pipeline2, pipeline3):
        for item in pipeline.run():
            assert len([x for x in item.data.keys() if x.startswith("text")]) == 2
            assert item.get_timing("reverser")
            assert item.get_timing("duplicator")
        assert pipeline.count == 10

    pipeline3 = (
        Pipeline()
        .set_source(RandomTextSource(10))
        .append("reverser", TextReverser(), concurrency=2)
        .append("error", ErrorStage(), concurrency=2, parallel=True)
        .set_error_manager(ErrorManager())
        .build()
    )
    pipeline4 = (
        Pipeline()
        .set_source(RandomTextSource(10))
        .append("reverser", TextReverser(), concurrency=2)
        .append_concurrently("error", ErrorStage, concurrency=2, parallel=True)
        .set_error_manager(ErrorManager())
        .build()
    )
    pipeline5 = (
        Pipeline()
        .set_source(RandomTextSource(10))
        .append_concurrently("reverser", TextReverser, concurrency=2)
        .append("error", ErrorStage(), concurrency=2, parallel=True)
        .set_error_manager(ErrorManager())
        .build()
    )
    for pipeline in (pipeline3, pipeline4, pipeline5):
        assert all(
            isinstance(c.error_manager, ErrorManager)
            for c in pipeline._containers.values()
        )
        for item in pipeline.run():
            assert item.has_soft_errors()
        assert pipeline.count == 10


# one core machines can have problems with this test
# TODO it has problems also on dual core machine
@pytest.mark.skip
def test_huge_run():
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(200))
        .append("reverser0", TextReverser(15000), concurrency=4, parallel=True)
        .append("reverser1", TextReverser(15000), concurrency=4, parallel=True)
        .append("reverser2", TextReverser(15000), concurrency=4, parallel=True)
        .append("duplicator", TextDuplicator(15000), concurrency=4, parallel=True)
        .build()
    )
    runner = pipeline.run()
    start_time = time.time()
    items = list(runner)
    elapsed1 = time.time() - start_time
    logger.debug("Time for strongly parallel: %s", elapsed1)
    _check(items, 200)
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(200))
        .append("reverser0", TextReverser(15000), concurrency=1, parallel=True)
        .append("reverser1", TextReverser(15000), concurrency=2, parallel=True)
        .append("reverser2", TextReverser(15000), concurrency=1, parallel=True)
        .append("duplicator", TextDuplicator(15000), concurrency=1, parallel=True)
        .build()
    )
    runner = pipeline.run()
    start_time = time.time()
    items = list(runner)
    elapsed2 = time.time() - start_time
    logger.debug("Time for mildly parallel: %s", elapsed2)
    _check(items, 200)
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(200))
        .append("reverser0", TextReverser(15000), concurrency=0)
        .append("reverser1", TextReverser(15000), concurrency=0)
        .append("reverser2", TextReverser(15000), concurrency=0)
        .append("duplicator", TextDuplicator(15000), concurrency=0)
        .build()
    )
    runner = pipeline.run()
    start_time = time.time()
    items = list(runner)
    elapsed3 = time.time() - start_time
    logger.debug("Time for sequential: %s", elapsed3)
    _check(items, 200)
    assert elapsed3 > elapsed1
    assert elapsed3 > elapsed2
    assert elapsed2 > elapsed1


def test_run_times():
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(10))
        .append("waster0", TimeWaster(0.2), concurrency=1)
        .append("waster1", TimeWaster(0.2), concurrency=1)
        .append("waster2", TimeWaster(0.2), concurrency=1)
        .append("waster3", TimeWaster(0.2), concurrency=1)
        .build()
    )
    start_time = time.time()
    items = list(pipeline.run())
    _check(items, 10)
    elapsed0 = time.time() - start_time
    logger.debug("Time for multi-threading: %s", elapsed0)
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(10))
        .append("waster0", TimeWaster(0.2), concurrency=1, parallel=True)
        .append("waster1", TimeWaster(0.2), concurrency=1, parallel=True)
        .append("waster2", TimeWaster(0.2), concurrency=1, parallel=True)
        .append("waster3", TimeWaster(0.2), concurrency=1, parallel=True)
        .build()
    )
    start_time = time.time()
    items = list(pipeline.run())
    _check(items, 10)
    elapsed1 = time.time() - start_time
    logger.debug("Time for multi-process: %s", elapsed1)
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(10))
        .append("waster0", TimeWaster(0.2), concurrency=0)
        .append("waster1", TimeWaster(0.2), concurrency=0)
        .append("waster2", TimeWaster(0.2), concurrency=0)
        .append("waster3", TimeWaster(0.2), concurrency=0)
        .build()
    )
    start_time = time.time()
    items = list(pipeline.run())
    _check(items, 10)
    elapsed2 = time.time() - start_time
    logger.debug("Time for sequential: %s", elapsed2)
    assert elapsed2 > elapsed0
    assert elapsed2 > elapsed1


def _check_item(item):
    assert item.data
    item.metadata["check"] = True


def test_single_items(items_generator_fx):
    pipeline = (
        get_pipeline()
        .append("reverser0", TextReverser())
        .append("reverser1", TextReverser())
        .append("reverser2", TextReverser())
        .append("duplicator", TextDuplicator())
        .build()
    )
    item = next(items_generator_fx)
    result = pipeline.process(copy.deepcopy(item))
    assert result.id == item.id
    assert result.data["text"] != item.data["text"]

    pipeline = (
        get_pipeline()
        .append_concurrently(
            "reverser0", TextReverser, kwargs={"cycles": 3}, concurrency=2
        )
        .append_concurrently("reverser1", TextReverser, args=[5], concurrency=0)
        .append("reverser2", TextReverser(), concurrency=1)
        .append("duplicator", TextDuplicator(), concurrency=2)
        .build()
    )
    with pytest.raises(ValueError):
        pipeline.get_item()
    item = next(items_generator_fx)
    pipeline.process_async(copy.deepcopy(item), callback=_check_item)
    pipeline.stop()
    result = pipeline.get_item()
    assert result.id == item.id
    assert result.metadata["check"]
    assert result.data["text"] != item.data["text"]

    pipeline = (
        get_pipeline()
        .append_concurrently("reverser0", TextReverser, concurrency=2, parallel=True)
        .append_concurrently(
            "reverser1", TextReverser, args=[9], concurrency=1, parallel=True
        )
        .append_concurrently("reverser2", TextReverser, concurrency=0)
        .append_concurrently(
            "duplicator", TextDuplicator, args=[10], concurrency=2, parallel=True
        )
        .build()
    )
    item = next(items_generator_fx)
    pipeline.process_async(item, callback=_check_item)
    pipeline.stop()
    result = pipeline.get_item()
    assert result.id == item.id
    assert result.metadata["check"]
    assert result.data["text"] != item.data["text"]

    pipeline = (
        get_pipeline()
        .append("reverser0", TextReverser(), concurrency=0)
        .append_concurrently("reverser1", TextReverser, concurrency=1)
        .append("duplicator", TextDuplicator(10), concurrency=0)
        .build()
    )
    item = next(items_generator_fx)
    pipeline.process_async(copy.deepcopy(item), callback=_check_item)
    pipeline.stop()
    result = pipeline.get_item()
    assert result.id == item.id
    assert result.metadata["check"]
    assert result.data["text"] == item.data["text"]

    pipeline = (
        get_pipeline()
        .append_concurrently("duplicator0", TextDuplicator)
        .append("reverser", TextReverser())
        .append_concurrently("duplicator1", TextDuplicator)
        .build()
    )
    item = next(items_generator_fx)
    pipeline.process_async(copy.deepcopy(item), callback=_check_item)
    pipeline.stop()
    result = pipeline.get_item()
    assert result.id == item.id
    assert result.metadata["check"]
    assert result.data["text"] != item.data["text"]

    pipeline = (
        get_pipeline()
        .append("reverser", TextReverser(11), concurrency=0)
        .append_concurrently("duplicator0", TextDuplicator, concurrency=0)
        .append("duplicator1", TextDuplicator(), concurrency=1)
        .build()
    )
    item = next(items_generator_fx)
    pipeline.process_async(copy.deepcopy(item), callback=_check_item)
    pipeline.stop()
    result = pipeline.get_item()
    assert result.id == item.id
    assert result.metadata["check"]
    assert result.data["text"] != item.data["text"]

    pipeline = (
        get_pipeline(max_init_workers=2)
        .append_concurrently(
            "reverser0", TextReverser, args=[20], concurrency=1, parallel=True
        )
        .append_concurrently(
            "reverser1", TextReverser, args=[20], concurrency=1, parallel=False
        )
        .append_concurrently(
            "duplicator", TextDuplicator, args=[20], concurrency=1, parallel=True
        )
        .build()
    )
    item = next(items_generator_fx)
    pipeline.process_async(item, callback=_check_item)
    pipeline.stop()
    result = pipeline.get_item()
    assert result.id == item.id
    assert result.metadata["check"]
    assert result.data["text"] == item.data["text"]
    assert len(result.data.keys()) > len(item.data.keys())


def test_concurrent_run_with_retryable_stages():
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(2))
        .append("reverser0", TextReverser(), concurrency=2)
        .append(
            "broken_stage1",
            CustomizableBrokenStage([ValueError]),
            concurrency=2,
            backoff=1,
            max_retries=2,
            retryable_errors=(ValueError,),
        )
        .append(
            "broken_stage2",
            CustomizableBrokenStage([TypeError]),
            concurrency=2,
            parallel=True,
            backoff=1,
            max_retries=1,
            retryable_errors=(TypeError,),
        )
        .append(
            "broken_stage3",
            CustomizableBrokenStage([IOError]),
            concurrency=1,
            parallel=True,
            backoff=0,
            max_retries=0,
            retryable_errors=(IOError,),
        )
        .build()
    )
    items = list(pipeline.run())
    for item in items:
        assert not item.has_critical_errors()
        soft_errors = list(item.soft_errors())
        soft_errors_broken_stage_1 = [
            err for err in soft_errors if err.get_stage() == "broken_stage1"
        ]
        soft_errors_broken_stage_2 = [
            err for err in soft_errors if err.get_stage() == "broken_stage2"
        ]
        soft_errors_broken_stage_3 = [
            err for err in soft_errors if err.get_stage() == "broken_stage3"
        ]
        assert len(soft_errors_broken_stage_1) == 3 and all(
            isinstance(err, RetryError) and isinstance(err.get_exception(), ValueError)
            for err in soft_errors_broken_stage_1
        )
        assert len(soft_errors_broken_stage_2) == 2 and all(
            isinstance(err, RetryError) and isinstance(err.get_exception(), TypeError)
            for err in soft_errors_broken_stage_2
        )
        assert len(soft_errors_broken_stage_3) == 1 and all(
            isinstance(err, RetryError) and isinstance(err.get_exception(), IOError)
            for err in soft_errors_broken_stage_3
        )


def test_broken_loop():
    def _get_pipeline():
        return (
            get_pipeline()
            .set_source(RandomTextSource(100))
            .append("reverser0", TextReverser(), concurrency=3, parallel=True)
            .append("reverser1", TextReverser(), concurrency=2, parallel=False)
            .append("duplicator", TextDuplicator(), concurrency=1, parallel=True)
            .build()
        )

    items = []
    for n, item in enumerate(_get_pipeline().run()):
        items.append(item)
        if n >= 9:
            break
    assert len(items) == 10
    items = []
    try:
        for n, item in enumerate(_get_pipeline().run()):
            items.append(item)
            if n >= 9:
                raise IOError
    except IOError:
        assert len(items) == 10


def test_logging(caplog):
    with caplog.at_level(logging.INFO):
        pipeline = (
            get_pipeline()
            .set_source(RandomTextSource(100))
            .append("logger1", Logger(), concurrency=2)
            .append("reverser0", TextReverser(), concurrency=1)
            .build()
        )
        list(pipeline.run())
        messages = [
            record for record in caplog.records if record.levelno == logging.INFO
        ]
        assert len(messages) == 100
        assert all("logging item " in record.msg.lower() for record in messages)
        caplog.clear()
        pipeline = (
            get_pipeline()
            .set_source(RandomTextSource(10))
            .append_concurrently("logger1", Logger, concurrency=1, parallel=True)
            .append("reverser0", TextReverser(), concurrency=1)
            .build()
        )
        list(pipeline.run())
        messages = [
            record for record in caplog.records if record.levelno == logging.INFO
        ]
        assert len(messages) == 10
        assert all("logging item " in record.msg.lower() for record in messages)
        assert any(
            "testing concurrent initialization" in record.msg.lower()
            for record in caplog.records
        )
        caplog.clear()
        pipeline = (
            get_pipeline()
            .set_source(RandomTextSource(100))
            .append("reverser1", TextReverser(), concurrency=2)
            .append("logger2", Logger(), concurrency=2, parallel=True)
            .append("reverser2", TextReverser())
            .build()
        )
        list(pipeline.run())
        messages = [
            record for record in caplog.records if record.levelno == logging.INFO
        ]
        assert len(messages) == 100
        assert all("logging item " in record.msg.lower() for record in messages)

        assert isinstance(
            _stage_initialization_with_logger(queue.Queue(), TextReverser, [], {}),
            TextReverser,
        )
        assert any(isinstance(h, QueueHandler) for h in logging.getLogger().handlers)

        with pytest.raises(AttributeError):
            get_pipeline().append(
                "reverser", TextReverser()
            ).build()._get_logs_receiver_queue()



import logging
import traceback

import pytest

from smartpipeline.error.exceptions import RetryError
from smartpipeline.error.handling import ErrorManager
from smartpipeline.pipeline import Pipeline
from tests.utils import (
    CriticalIOErrorStage,
    CustomException,
    CustomizableBrokenStage,
    ErrorStage,
    ExceptionStage,
    RandomTextSource,
    TextDuplicator,
    TextReverser,
    get_pipeline,
)

__author__ = "Giacomo Berardi <giacbrd.com>"

logger = logging.getLogger(__name__)


def test_run():
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(10))
        .append("reverser", TextReverser())
        .append("duplicator", TextDuplicator())
        .build()
    )
    assert pipeline.name.startswith("Pipeline-")
    assert str(pipeline._source_container)
    assert isinstance(pipeline.get_stage("duplicator"), TextDuplicator)
    for container in pipeline._containers.values():
        assert container.name
        assert str(container)
    for item in pipeline.run():
        assert len([x for x in item.data.keys() if x.startswith("text")]) == 2
        assert item.get_timing("reverser")
        assert item.get_timing("duplicator")
    assert pipeline.count == 10
    with pytest.raises(ValueError):
        Pipeline().append("reverser", TextReverser()).append("reverser", TextReverser())


def test_errors(caplog):
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(10))
        .append("reverser", TextReverser())
        .append("error", ErrorStage())
        .build()
    )
    error_manager = ErrorManager()
    pipeline.set_error_manager(error_manager)
    assert all(c.error_manager == error_manager for c in pipeline._containers.values())
    for item in pipeline.run():
        assert item.has_soft_errors()
        assert item.get_timing("reverser")
        assert item.get_timing("error")
        error = next(item.soft_errors())
        assert error.get_exception() is None
        assert str(error) == "test pipeline error"
    assert all(
        "has generated an error" in record.msg.lower()
        and "stage error" == str(record.args[0]).lower()
        for record in caplog.records
        if record.levelno == logging.ERROR
    )
    assert pipeline.count == 10
    pipeline = (
        get_pipeline()
        .set_error_manager(ErrorManager())
        .set_source(RandomTextSource(10))
        .append("reverser", TextReverser())
        .append("error", ErrorStage())
        .append("duplicator", TextDuplicator())
        .build()
    )
    caplog.clear()
    for item in pipeline.run():
        assert item.has_soft_errors()
        assert item.get_timing("reverser")
        assert item.get_timing("duplicator")
        assert any(k.startswith("text_") for k in item.data.keys())
        assert item.get_timing("error")
        error = next(item.soft_errors())
        assert error.get_exception() is None
        assert str(error) == "test pipeline error"
    assert all(
        "has generated an error" in record.msg.lower()
        and "stage error" == str(record.args[0]).lower()
        for record in caplog.records
        if record.levelno == logging.ERROR
    )
    assert pipeline.count == 10
    pipeline = (
        get_pipeline()
        .set_error_manager(ErrorManager())
        .set_source(RandomTextSource(10))
        .append("reverser", TextReverser())
        .append("error1", CriticalIOErrorStage())
        .build()
    )
    for item in pipeline.run():
        assert item.has_critical_errors()
        assert item.get_timing("reverser")
        assert item.get_timing("error1")
        for error in item.critical_errors():
            assert isinstance(error.get_exception(), IOError)
            assert str(error) == "test pipeline critical IO error"
    pipeline = (
        get_pipeline()
        .set_error_manager(ErrorManager())
        .set_source(RandomTextSource(10))
        .append("reverser", TextReverser())
        .append("error1", ExceptionStage())
        .append("error2", ErrorStage())
        .build()
    )
    caplog.clear()
    for item in pipeline.run():
        assert item.has_critical_errors()
        assert item.get_timing("reverser")
        assert item.get_timing("error1")
        assert not item.get_timing("error2")
        for error in item.critical_errors():
            assert isinstance(error.get_exception(), Exception)
            assert (
                str(error) == "test exception"
                and str(error.get_exception()) == "test exception"
                and str(error) != "test pipeline error"
            )
    assert all(
        "has generated an error" in record.msg.lower()
        and "stage error1" == str(record.args[0]).lower()
        for record in caplog.records
        if record.levelno == logging.ERROR
    )
    assert pipeline.count == 10
    with pytest.raises(Exception):
        pipeline = (
            get_pipeline()
            .set_source(RandomTextSource(10))
            .append("reverser", TextReverser())
            .append("error", ExceptionStage())
            .build()
        )
        try:
            for _ in pipeline.run():
                pass
        except Exception:
            assert 'Exception("test exception")' in traceback.format_exc()
            raise
        assert pipeline.count == 1
    pipeline = (
        get_pipeline()
        .set_error_manager(ErrorManager().no_skip_on_critical_error())
        .set_source(RandomTextSource(10))
        .append("reverser1", TextReverser())
        .append("error", ExceptionStage())
        .append("reverser2", TextReverser())
        .build()
    )
    for item in pipeline.run():
        assert item.get_timing("reverser1")
        assert item.get_timing("error")
        assert item.get_timing("reverser2")
    assert pipeline.count == 10


def test_build():
    """Test the build pattern with non-conventional order of the method calls"""
    pipeline1 = (
        Pipeline()
        .append("reverser", TextReverser())
        .append("duplicator", TextDuplicator())
        .set_source(RandomTextSource(10))
        .build()
    )
    pipeline2 = (
        Pipeline()
        .append("reverser", TextReverser())
        .set_source(RandomTextSource(10))
        .append("duplicator", TextDuplicator())
        .build()
    )
    for pipeline in (pipeline1, pipeline2):
        for item in pipeline.run():
            assert len([x for x in item.data.keys() if x.startswith("text")]) == 2
            assert item.get_timing("reverser")
            assert item.get_timing("duplicator")
        assert pipeline.count == 10

    pipeline3 = (
        Pipeline()
        .set_source(RandomTextSource(10))
        .append("reverser", TextReverser())
        .set_error_manager(ErrorManager())
        .append("error", ErrorStage())
        .build()
    )
    pipeline4 = (
        Pipeline()
        .set_source(RandomTextSource(10))
        .append("reverser", TextReverser())
        .append("error", ErrorStage())
        .set_error_manager(ErrorManager())
        .build()
    )
    for pipeline in (
        pipeline3,
        pipeline4,
    ):
        assert all(
            isinstance(c.error_manager, ErrorManager)
            for c in pipeline._containers.values()
        )
        for item in pipeline.run():
            assert item.has_soft_errors()
        assert pipeline.count == 10


def test_retryable_stage(items_generator_fx):
    # default behaviour
    item = next(items_generator_fx)
    pipeline = (
        Pipeline()
        .append("reverser0", TextReverser())
        .append("broken_stage", CustomizableBrokenStage([ValueError]))
        .append("reverser1", TextReverser())
    )
    item = pipeline.process(item)
    assert not item.has_soft_errors()  # no soft errors
    critical_errors = list(item.critical_errors())
    assert len(critical_errors) == 1 and isinstance(
        critical_errors[0].get_exception(), ValueError
    )

    # not defining any error to which retry on, so the pipeline behaviour should not change
    item = next(items_generator_fx)
    pipeline = (
        Pipeline()
        .append("reverser0", TextReverser())
        .append(
            "broken_stage",
            CustomizableBrokenStage([ValueError]),
            backoff=1,
            max_retries=1,
        )
        .append("reverser1", TextReverser())
        .build()
    )
    item = pipeline.process(item)
    assert not item.has_soft_errors()  # no soft errors
    critical_errors = list(item.critical_errors())
    assert len(critical_errors) == 1 and isinstance(
        critical_errors[0].get_exception(), ValueError
    )

    # defining some retryable_errors for which the stage is force to try the reprocessing of the item
    item = next(items_generator_fx)
    pipeline = (
        Pipeline()
        .append("reverser0", TextReverser())
        .append(
            "broken_stage",
            CustomizableBrokenStage([ValueError]),
            backoff=1,
            max_retries=1,
            retryable_errors=(ValueError,),
        )
        .append("reverser1", TextReverser())
        .build()
    )
    item = pipeline.process(item)
    assert not item.has_critical_errors()
    soft_errors = list(item.soft_errors())
    assert len(soft_errors) == 2 and all(
        isinstance(err, RetryError) and isinstance(err.get_exception(), ValueError)
        for err in soft_errors
    )

    # 0 retries
    item = next(items_generator_fx)
    pipeline = (
        Pipeline()
        .append("reverser0", TextReverser())
        .append(
            "broken_stage",
            CustomizableBrokenStage([ValueError]),
            backoff=0,
            max_retries=0,
            retryable_errors=(ValueError,),
        )
        .append("reverser1", TextReverser())
        .build()
    )
    item = pipeline.process(item)
    assert not item.has_critical_errors()
    soft_errors = list(item.soft_errors())
    assert len(soft_errors) == 1 and all(
        isinstance(err, RetryError) and isinstance(err.get_exception(), ValueError)
        for err in soft_errors
    )

    # mixing retryable and critical errors
    item = next(items_generator_fx)
    pipeline = (
        Pipeline()
        .append("reverser0", TextReverser())
        .append(
            "broken_stage0",
            CustomizableBrokenStage([CustomException]),
            backoff=1,
            max_retries=1,
            retryable_errors=(ValueError, CustomException),
        )
        .append("reverser1", TextReverser())
        .append(
            "broken_stage1",
            CustomizableBrokenStage([CustomException]),
            backoff=1,
            max_retries=2,
            retryable_errors=(ValueError,),
        )
        .build()
    )

    item = pipeline.process(item)
    critical_errors = list(item.critical_errors())
    assert len(critical_errors) == 1 and (
        critical_errors[0].get_exception(),
        CustomException,
    )
    soft_errors = list(item.soft_errors())
    assert len(soft_errors) == 2 and all(
        isinstance(err, RetryError) and isinstance(err.get_exception(), CustomException)
        for err in soft_errors
    )
    with pytest.raises(ValueError):
        Pipeline().append(
            "broken_stage",
            CustomizableBrokenStage([CustomException]),
            backoff=1,
            max_retries=1,
            retryable_errors=[ValueError, CustomException],
        )
    with pytest.raises(ValueError):
        Pipeline().append(
            "broken_stage",
            CustomizableBrokenStage([CustomException]),
            backoff="1.2",
            max_retries=1,
            retryable_errors=(ValueError, CustomException),
        )
    with pytest.raises(ValueError):
        Pipeline().append(
            "broken_stage",
            CustomizableBrokenStage([CustomException]),
            backoff=1,
            max_retries=2.5,
            retryable_errors=(ValueError, CustomException),
        )


def test_retry_on_different_errors(items_generator_fx):
    item = next(items_generator_fx)
    pipeline = (
        Pipeline()
        .append("reverser0", TextReverser())
        .append(
            "broken_stage",
            CustomizableBrokenStage([CustomException, ValueError]),
            backoff=0,
            max_retries=2,
            retryable_errors=(CustomException, ValueError),
        )
        .append("reverser1", TextReverser())
        .build()
    )
    item = pipeline.process(item)
    soft_errors = list(item.soft_errors())
    assert len(soft_errors) == 3
    assert all(
        isinstance(err, RetryError) and err.get_stage() == "broken_stage"
        for err in soft_errors
    )
    assert isinstance(soft_errors[0].get_exception(), CustomException) and isinstance(
        soft_errors[2].get_exception(), CustomException
    )
    assert isinstance(soft_errors[1].get_exception(), ValueError)


def test_exponential_backoff_retry_strategy(items_generator_fx):
    item = next(items_generator_fx)
    pipeline = (
        Pipeline()
        .append("reverser0", TextReverser())
        .append(
            "broken_stage",
            CustomizableBrokenStage([CustomException]),
            backoff=1,
            max_retries=0,
            retryable_errors=(CustomException,),
        )
        .append("reverser1", TextReverser())
        .build()
    )
    item = pipeline.process(item)
    assert item.get_timing("broken_stage") < 1
    item = next(items_generator_fx)
    pipeline = (
        Pipeline()
        .append("reverser0", TextReverser())
        .append(
            "broken_stage",
            CustomizableBrokenStage([CustomException]),
            backoff=1,
            max_retries=3,
            retryable_errors=(CustomException,),
        )
        .append("reverser1", TextReverser())
        .build()
    )
    item = pipeline.process(item)
    assert 15 <= item.get_timing("broken_stage") <= 16
    soft_errors = list(item.soft_errors())
    assert len(soft_errors) == 4
    assert all(
        isinstance(err, RetryError) and isinstance(err.get_exception(), CustomException)
        for err in soft_errors
    )



import copy
import logging

import pytest

from smartpipeline.error.exceptions import RetryError
from smartpipeline.error.handling import ErrorManager
from smartpipeline.pipeline import Pipeline
from tests.utils import (
    BatchErrorStage,
    BatchExceptionStage,
    BatchTextDuplicator,
    BatchTextReverser,
    CustomException,
    CustomizableBrokenBatchStage,
    ErrorStage,
    ExceptionStage,
    RandomTextSource,
    TextDuplicator,
    TextReverser,
    get_pipeline,
)

__author__ = "Giacomo Berardi <giacbrd.com>"

logger = logging.getLogger(__name__)


def test_run():
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(35))
        .append("reverser", BatchTextReverser())
        .append("duplicator", TextDuplicator())
        .append("batch_duplicator", BatchTextDuplicator())
        .build()
    )
    for item in pipeline.run():
        assert len([x for x in item.data.keys() if x.startswith("text")]) == 3
        assert item.get_timing("reverser")
        assert item.get_timing("duplicator")
    assert pipeline.count == 35
    for _ in range(3):
        assert pipeline._containers["batch_duplicator"].get_processed() is None
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(40))
        .append("reverser", TextReverser())
        .append("duplicator", TextDuplicator())
        .append("batch_duplicator", BatchTextDuplicator(check_batch=True))
        .build()
    )
    for _ in pipeline.run():
        pass
    assert pipeline.count == 40


def test_run_different_sizes():
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(2))
        .append("reverser", BatchTextReverser(size=4))
        .append("duplicator", BatchTextDuplicator(size=20))
        .build()
    )
    n = 0
    for n, item in enumerate(pipeline.run()):
        assert len([x for x in item.data.keys() if x.startswith("text")]) == 2
        assert item.get_timing("reverser")
        assert item.get_timing("duplicator")
    assert n == 1 == pipeline.count - 1
    pipeline = (
        get_pipeline()
        .set_source(RandomTextSource(6))
        .append("reverser", BatchTextReverser(size=1))
        .append("duplicator", BatchTextDuplicator(size=20))
        .build()
    )
    for n, item in enumerate(pipeline.run()):
        assert len([x for x in item.data.keys() if x.startswith("text")]) == 2
        assert item.get_timing("reverser")
        assert item.get_timing("duplicator")
    assert n == 5 == pipeline.count - 1


def test_errors(caplog):
    pipeline = (
        get_pipeline()
        .set_error_manager(ErrorManager())
        .set_source(RandomTextSource(22))
        .append("reverser", BatchTextReverser(size=5))
        .append("error", ErrorStage())
        .build()
    )
    for item in pipeline.run():
        assert item.has_soft_errors()
        assert item.get_timing("reverser")
        assert item.get_timing("error")
        error = next(item.soft_errors())
        assert error.get_exception() is None
        assert str(error) == "test pipeline error"
    assert pipeline.count == 22
    assert all(
        "has generated an error" in record.msg.lower()
        for record in caplog.records
        if record.levelno == logging.ERROR
    )
    pipeline = (
        get_pipeline()
        .set_error_manager(ErrorManager())
        .set_source(RandomTextSource(28))
        .append("reverser", BatchTextReverser(size=8))
        .append("error", ErrorStage())
        .append("duplicator", BatchTextDuplicator(size=5))
        .build()
    )
    caplog.clear()
    for item in pipeline.run():
        assert item.has_soft_errors()
        assert item.get_timing("reverser")
        assert item.get_timing("duplicator")
        assert any(k.startswith("text_") for k in item.data.keys())
        assert item.get_timing("error")
        error = next(item.soft_errors())
        assert error.get_exception() is None
        assert str(error) == "test pipeline error"
    assert pipeline.count == 28
    assert any(caplog.records)
    pipeline = (
        get_pipeline()
        .set_error_manager(ErrorManager())
        .set_source(RandomTextSource(10))
        .append("reverser", BatchTextReverser(size=3))
        .append("error1", ExceptionStage())
        .append("error2", ErrorStage())
        .build()
    )
    caplog.clear()
    for item in pipeline.run():
        assert item.has_critical_errors()
        assert item.get_timing("reverser")
        assert item.get_timing("error1") >= 0.0003
        assert not item.get_timing("error2")
        for error in item.critical_errors():
            assert isinstance(error.get_exception(), Exception)
            assert (
                str(error.get_exception()) == "test exception"
                and str(error) != "test pipeline error"
            )
    assert pipeline.count == 10
    assert any(caplog.records)
    with pytest.raises(Exception):
        pipeline = (
            get_pipeline()
            .set_source(RandomTextSource(10))
            .append("reverser", BatchTextReverser(size=4))
            .append("error", ExceptionStage())
            .build()
        )
        for _ in pipeline.run():
            pass
        assert pipeline.count == 1
    pipeline = (
        get_pipeline()
        .set_error_manager(ErrorManager().no_skip_on_critical_error())
        .set_source(RandomTextSource(10))
        .append("reverser1", BatchTextReverser(size=1))
        .append("error", ExceptionStage())
        .append("reverser2", BatchTextReverser(size=4))
        .build()
    )
    for item in pipeline.run():
        assert item.get_timing("reverser1")
        assert item.get_timing("error")
        assert item.get_timing("reverser2")
    assert pipeline.count == 10


def test_batch_errors(caplog):
    pipeline = (
        get_pipeline()
        .set_error_manager(ErrorManager())
        .set_source(RandomTextSource(22))
        .append("reverser", BatchTextReverser(size=5))
        .append("error", BatchErrorStage(size=3))
        .build()
    )
    for item in pipeline.run():
        assert item.has_soft_errors()
        assert item.get_timing("reverser")
        assert item.get_timing("error")
        error = next(item.soft_errors())
        assert error.get_exception() is None
        assert str(error) == "test pipeline error"
    assert pipeline.count == 22
    assert any(caplog.records)
    pipeline = (
        get_pipeline()
        .set_error_manager(ErrorManager())
        .set_source(RandomTextSource(28))
        .append("reverser", BatchTextReverser(size=8))
        .append("error", BatchErrorStage(size=7))
        .append("duplicator", BatchTextDuplicator(size=5))
        .build()
    )
    for item in pipeline.run():
        assert item.has_soft_errors()
        assert item.get_timing("reverser")
        assert item.get_timing("duplicator")
        assert any(k.startswith("text_") for k in item.data.keys())
        assert item.get_timing("error")
        error = next(item.soft_errors())
        assert error.get_exception() is None
        assert str(error) == "test pipeline error"
    assert pipeline.count == 28
    assert any(caplog.records)
    pipeline = (
        get_pipeline()
        .set_error_manager(ErrorManager())
        .set_source(RandomTextSource(10))
        .append("reverser", BatchTextReverser(size=3))
        .append("error1", BatchExceptionStage(size=7))
        .append("error2", BatchErrorStage(size=1))
        .build()
    )
    for item in pipeline.run():
        assert item.has_critical_errors()
        assert item.get_timing("reverser")
        assert item.get_timing("error1") >= 0.0003
        assert not item.get_timing("error2")
        for error in item.critical_errors():
            assert isinstance(error.get_exception(), Exception)
            assert (
                str(error.get_exception()) == "test exception"
                and str(error) != "test pipeline error"
            )
    assert pipeline.count == 10
    assert any(caplog.records)
    with pytest.raises(Exception):
        pipeline = (
            get_pipeline()
            .set_source(RandomTextSource(10))
            .append("reverser", BatchTextReverser(size=4))
            .append("error", BatchExceptionStage(size=3))
            .build()
        )
        for _ in pipeline.run():
            pass
        assert pipeline.count == 1


def test_retryable_batch_stages(items_generator_fx):
    pipeline = (
        Pipeline()
        .set_source(RandomTextSource(10))
        .append("reverser0", TextReverser())
        .append(
            "broken_batch_stage1",
            CustomizableBrokenBatchStage(
                size=10, timeout=5, exceptions_to_raise=[CustomException, ValueError]
            ),
            backoff=0.01,
            max_retries=1,
            retryable_errors=(CustomException, ValueError),
        )
        .append(
            "broken_batch_stage2",
            CustomizableBrokenBatchStage(
                size=10, timeout=5, exceptions_to_raise=[IOError]
            ),
            backoff=0,
            max_retries=0,
            retryable_errors=(IOError,),
        )
        .append("reverser1", TextReverser())
        .build()
    )
    item = next(items_generator_fx)
    pipeline.process(copy.deepcopy(item))
    items = list(pipeline.run())
    for item in items:
        assert not item.has_critical_errors()
        assert all(
            isinstance(err, RetryError)
            and isinstance(err.get_exception(), (CustomException, ValueError, IOError))
            and err.get_stage() == "broken_batch_stage1"
            or err.get_stage() == "broken_batch_stage2"
            for err in item.soft_errors()
        )
